/** \file
 * Интерфейс прикладного программирования на языке C++ к библиотеке криптографических алгоритмов ГОСТ.
 */

#ifndef GOSTCRYPTO_HPP
#define GOSTCRYPTO_HPP

#include "gostcrypto.h"
#include <type_traits>
#include <array>
#include <vector>

#if __cplusplus < 201402L
template <std::size_t Len, std::size_t Align>
using aligned_storage_t = typename std::aligned_storage<Len, Align>::type;
#else
template <std::size_t Len, std::size_t Align>
using aligned_storage_t = std::aligned_storage_t<Len, Align>;
#endif

/**
 * Блочный шифр с длиной блока 128 бит в режиме работы блочных шифров, реализующим
 * аутентифицирующее шифрование с присоединёнными данными (AEAD). Определен в Р 1323565.1.026–2019.
 *
 * \see #KuznyechikMgm
 */
class KuznyechikMgm_t
{
public:
  KuznyechikMgm_t(const KuznyechikMgm_t &) = default;
  KuznyechikMgm_t &operator=(const KuznyechikMgm_t &) = default;
  KuznyechikMgm_t(KuznyechikMgm_t &&) = default;
  KuznyechikMgm_t &operator=(KuznyechikMgm_t &&) = default;

  /**
   * @brief Создать контекст AEAD.
   *
   * Создаёт контекст AEAD с использованием ключа `key` в заданной области памяти.
   *
   * @param key — ключ
   */
  explicit KuznyechikMgm_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key) noexcept
      : data_{}
  {
    kuznyechik_mgm_new(*ptr(), key.data());
  }

  /**
   * @brief Зашифровать сообщение с присоединёнными данными.
   *
   * Зашифровывает сообщение на месте с присоединёнными данными и копирует соответствующую
   * имитовставку в массив `tag`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param nonce — синхропосылка как массив байт длиной #KUZNYECHIK_MGM_NONCE_SIZE. Старший
   * бит `nonce[0]` должен быть равен нулю.
   * @param aad — присоединённые данные
   * @param aad_len — длина присоединённых данных в байтах
   * @param buf — буфер, содержащий подлежащие шифрованию данные
   * @param buf_len — длина буфера (в байтах)
   * @param tag — имитовставка как массив байт длиной #KUZNYECHIK_MGM_TAG_SIZE
   *
   * @returns 0, в случае успеха, -1, в случае ошибки
   */
  intptr_t encrypt(const uint8_t nonce[KUZNYECHIK_MGM_NONCE_SIZE],
                   const uint8_t *aad,
                   uintptr_t aad_len,
                   uint8_t *buf,
                   uintptr_t buf_len,
                   uint8_t tag[KUZNYECHIK_MGM_TAG_SIZE]) const noexcept
  {
    return kuznyechik_mgm_encrypt(*ptr(), nonce, aad, aad_len, buf, buf_len, tag);
  }
  /**
   * @brief Зашифровать сообщение с присоединёнными данными.
   *
   * Зашифровывает сообщение на месте с присоединёнными данными и копирует соответствующую
   * имитовставку в массив `tag`.
   *
   * @param nonce — синхропосылка. Старший бит `nonce[0]` должен быть равен нулю.
   * @param aad — присоединённые данные
   * @param buf — буфер, содержащий подлежащие шифрованию данные
   * @param tag — имитовставка
   *
   * @returns 0, в случае успеха, -1, в случае ошибки
   */
  intptr_t encrypt(const std::array<uint8_t, KUZNYECHIK_MGM_NONCE_SIZE> &nonce,
                   const std::vector<uint8_t> &aad,
                   std::vector<uint8_t> &buf,
                   std::array<uint8_t, KUZNYECHIK_MGM_TAG_SIZE> &tag) const noexcept
  {
    return kuznyechik_mgm_encrypt(*ptr(), nonce.data(), aad.data(), aad.size(), buf.data(), buf.size(), tag.data());
  }

  /**
   * @brief Расшифровать сообщение с присоединёнными данными.
   *
   * Расшифровывает сообщение на месте с присоединёнными данными с использованием данной
   * имитовставки `tag`. Возвращает ошибку в случае, если имитовставка не соответствует
   * шифротексту.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param nonce — синхропосылка как массив байт длиной #KUZNYECHIK_MGM_NONCE_SIZE
   * @param aad — присоединённые данные
   * @param aad_len — длина присоединённых данных в байтах
   * @param buf — буфер, содержащий подлежащие расшифрованию данные
   * @param buf_len — длина буфера (в байтах)
   * @param tag — имитовставка как массив байт длиной #KUZNYECHIK_MGM_TAG_SIZE
   *
   * @returns 0, в случае успеха, -1, в случае ошибки
   */
  intptr_t decrypt(const uint8_t nonce[KUZNYECHIK_MGM_NONCE_SIZE],
                   const uint8_t *aad,
                   uintptr_t aad_len,
                   uint8_t *buf,
                   uintptr_t buf_len,
                   const uint8_t tag[KUZNYECHIK_MGM_TAG_SIZE]) const noexcept
  {
    return kuznyechik_mgm_decrypt(*ptr(), nonce, aad, aad_len, buf, buf_len, tag);
  }

  /**
   * @brief Расшифровать сообщение с присоединёнными данными.
   *
   * Расшифровывает сообщение на месте с присоединёнными данными с использованием данной
   * имитовставки `tag`. Возвращает ошибку в случае, если имитовставка не соответствует
   * шифротексту.
   *
   * @param nonce — синхропосылка как массив байт длиной #KUZNYECHIK_MGM_NONCE_SIZE
   * @param aad — присоединённые данные
   * @param buf — буфер, содержащий подлежащие расшифрованию данные
   * @param tag — имитовставка
   *
   * @returns 0, в случае успеха, -1, в случае ошибки
   */
  intptr_t decrypt(const std::array<uint8_t, KUZNYECHIK_MGM_NONCE_SIZE> &nonce,
                   const std::vector<uint8_t> &aad,
                   std::vector<uint8_t> &buf,
                   const std::array<uint8_t, KUZNYECHIK_MGM_TAG_SIZE> &tag) const noexcept
  {
    return kuznyechik_mgm_decrypt(*ptr(), nonce.data(), aad.data(), aad.size(), buf.data(), buf.size(), tag.data());
  }

  ~KuznyechikMgm_t()
  {
    kuznyechik_mgm_drop(*ptr());
  }

private:
  KuznyechikMgm *ptr() noexcept
  {
    return reinterpret_cast<KuznyechikMgm *>(&data_);
  }

  KuznyechikMgm *ptr() const noexcept
  {
    return const_cast<KuznyechikMgm *>(reinterpret_cast<const KuznyechikMgm *>(&data_));
  }

  aligned_storage_t<KUZNYECHIK_MGM_SIZE, KUZNYECHIK_MGM_ALIGNMENT> data_;
};

/**
 * Блочный шифр с длиной блока 128 бит в режиме простой замены с зацеплением
 * с размером синхропосылки 16 байт. Определен в ГОСТ 34.13-2015 и ГОСТ 34.13-2018.
 *
 * \see #KuznyechikCbc
 */
class KuznyechikCbc_t
{
public:
  KuznyechikCbc_t(const KuznyechikCbc_t &) = default;
  KuznyechikCbc_t &operator=(const KuznyechikCbc_t &) = default;
  KuznyechikCbc_t(KuznyechikCbc_t &&) = default;
  KuznyechikCbc_t &operator=(KuznyechikCbc_t &&) = default;

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
   * ключа `key` и синхропосылки `nonce`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   * @param nonce — ненулевой указатель на массив длинной #KUZNYECHIK_CBC_NONCE_SIZE байт
   */
  explicit KuznyechikCbc_t(const uint8_t key[SYMMETRIC_KEY_SIZE], const uint8_t nonce[KUZNYECHIK_CBC_NONCE_SIZE]) noexcept
      : data_{}
  {
    kuznyechik_cbc_new(*ptr(), reinterpret_cast<const uint8_t *>(key), reinterpret_cast<const uint8_t *>(nonce));
  }

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
   * ключа `key` и синхропосылки `nonce`.
   *
   * @param key — ключ
   * @param nonce — синхропосылка
   */
  explicit KuznyechikCbc_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key, const std::array<uint8_t, KUZNYECHIK_CBC_NONCE_SIZE> &nonce) noexcept
      : data_{}
  {
    kuznyechik_cbc_new(*ptr(), key.data(), nonce.data());
  }

  ~KuznyechikCbc_t() noexcept
  {
    kuznyechik_cbc_drop(*ptr());
  }

  /**
   * @brief Зашифровать массив блоков.
   *
   * Шифрует блоки данных на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков. Блок представляет собой
   * последовательность байт длиной #KUZNYECHIK_BLOCK_SIZE.
   * @param blocks_len — длина массива блоков (в блоках)
   */
  void encrypt_blocks(uint8_t (*blocks)[KUZNYECHIK_BLOCK_SIZE], uintptr_t blocks_len) noexcept
  {
    kuznyechik_cbc_encrypt_blocks(*ptr(), reinterpret_cast<uint8_t *>(blocks), blocks_len);
  }

  /**
   * @brief Расшифровать массив блоков.
   *
   * Расшифровывает блоки данных на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков. Блок представляет собой
   * последовательность байт длиной #KUZNYECHIK_BLOCK_SIZE.
   * @param blocks_len — длина массива блоков (в блоках)
   */
  void decrypt_blocks(uint8_t (*blocks)[KUZNYECHIK_BLOCK_SIZE], uintptr_t blocks_len) noexcept
  {
    kuznyechik_cbc_decrypt_blocks(*ptr(), reinterpret_cast<uint8_t *>(blocks), blocks_len);
  }

  /**
   * @brief Зашифровать сообщение.
   *
   * Зашифровывает сообщение на месте.
   *
   * Байты `0..pos` буфера `buf` используются как сообщение, когда же байты `pos..buf_len`
   * используются как зарезервированное место для дополнения.
   *
   * Возвращает ошибку в случае, когда место для дополнения недостаточно велико.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   * @param pos — индекс байта, с которого начинается дополнение в буфере `buf`
   *
   * @returns Длину шифротекста, в случае успеха или -1, в случае ошибки
   */
  intptr_t encrypt_pad(uint8_t *buf, uintptr_t buf_len, uintptr_t pos) const noexcept
  {
    auto copy = *this;
    return kuznyechik_cbc_encrypt_pad(*copy.ptr(), buf, buf_len, pos);
  }

  /**
   * @brief Расшифровать сообщение.
   *
   * Расшифровывает сообщение на месте.
   *
   * Возвращает ошибку в случае, когда длина буфера не делится на #KUZNYECHIK_BLOCK_SIZE или
   * расшифрованное сообщение имеет некорректное дополнение.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   *
   * @returns Длину сообщения, в случае успеха или -1, в случае ошибки
   */
  intptr_t decrypt_pad(uint8_t *buf, uintptr_t buf_len) const noexcept
  {
    auto copy = *this;
    return kuznyechik_cbc_decrypt_pad(*copy.ptr(), buf, buf_len);
  }

private:
  KuznyechikCbc *ptr() noexcept
  {
    return reinterpret_cast<KuznyechikCbc *>(&data_);
  }

  KuznyechikCbc *ptr() const noexcept
  {
    return const_cast<KuznyechikCbc *>(reinterpret_cast<const KuznyechikCbc *>(&data_));
  }

  aligned_storage_t<KUZNYECHIK_CBC_SIZE, KUZNYECHIK_CBC_ALIGNMENT> data_;
};

/**
 * Блочный шифр с длиной блока 64 бит в режиме простой замены с зацеплением
 * с размером синхропосылки 8 байт. Определен в ГОСТ 34.13-2015 и ГОСТ 34.13-2018.
 *
 * \see #MagmaCbc
 */
class MagmaCbc_t
{
public:
  MagmaCbc_t(const MagmaCbc_t &) = default;
  MagmaCbc_t &operator=(const MagmaCbc_t &) = default;
  MagmaCbc_t(MagmaCbc_t &&) = default;
  MagmaCbc_t &operator=(MagmaCbc_t &&) = default;

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
   * ключа `key` и синхропосылки `nonce`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   * @param nonce — ненулевой указатель на массив длинной #MAGMA_CBC_NONCE_SIZE байт
   */
  explicit MagmaCbc_t(const uint8_t key[SYMMETRIC_KEY_SIZE], const uint8_t nonce[MAGMA_CBC_NONCE_SIZE]) noexcept
      : data_{}
  {
    magma_cbc_new(*ptr(), reinterpret_cast<const uint8_t *>(key), reinterpret_cast<const uint8_t *>(nonce));
  }

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
   * ключа `key` и синхропосылки `nonce`.
   *
   * @param key — ключ
   * @param nonce — синхропосылка
   */
  explicit MagmaCbc_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key, const std::array<uint8_t, MAGMA_CBC_NONCE_SIZE> &nonce) noexcept
      : data_{}
  {
    magma_cbc_new(*ptr(), key.data(), nonce.data());
  }

  ~MagmaCbc_t() noexcept
  {
    magma_cbc_drop(*ptr());
  }

  /**
   * @brief Зашифровать массив блоков.
   *
   * Шифрует блоки данных на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков. Блок представляет собой
   * последовательность байт длиной #MAGMA_BLOCK_SIZE.
   * @param blocks_len — длина массива блоков (в блоках)
   */
  void encrypt_blocks(uint8_t (*blocks)[MAGMA_BLOCK_SIZE], uintptr_t blocks_len) noexcept
  {
    magma_cbc_encrypt_blocks(*ptr(), reinterpret_cast<uint8_t *>(blocks), blocks_len);
  }

  /**
   * @brief Расшифровать массив блоков.
   *
   * Расшифровывает блоки данных на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков. Блок представляет собой
   * последовательность байт длиной #MAGMA_BLOCK_SIZE.
   * @param blocks_len — длина массива блоков (в блоках)
   */
  void decrypt_blocks(uint8_t (*blocks)[MAGMA_BLOCK_SIZE], uintptr_t blocks_len) noexcept
  {
    magma_cbc_decrypt_blocks(*ptr(), reinterpret_cast<uint8_t *>(blocks), blocks_len);
  }

  /**
   * @brief Зашифровать сообщение.
   *
   * Зашифровывает сообщение на месте.
   *
   * Байты `0..pos` буфера `buf` используются как сообщение, когда же байты `pos..buf_len`
   * используются как зарезервированное место для дополнения.
   *
   * Возвращает ошибку в случае, когда место для дополнения недостаточно велико.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   * @param pos — индекс байта, с которого начинается дополнение в буфере `buf`
   *
   * @returns Длину шифротекста, в случае успеха или -1, в случае ошибки
   */
  intptr_t encrypt_pad(uint8_t *buf, uintptr_t buf_len, uintptr_t pos) const noexcept
  {
    auto copy = *this;
    return magma_cbc_encrypt_pad(*copy.ptr(), buf, buf_len, pos);
  }

  /**
   * @brief Расшифровать сообщение.
   *
   * Расшифровывает сообщение на месте.
   *
   * Возвращает ошибку в случае, когда длина буфера не делится на #MAGMA_BLOCK_SIZE или
   * расшифрованное сообщение имеет некорректное дополнение.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   *
   * @returns Длину сообщения, в случае успеха или -1, в случае ошибки
   */
  intptr_t decrypt_pad(uint8_t *buf, uintptr_t buf_len) const noexcept
  {
    auto copy = *this;
    return magma_cbc_decrypt_pad(*copy.ptr(), buf, buf_len);
  }

private:
  MagmaCbc *ptr() noexcept
  {
    return reinterpret_cast<MagmaCbc *>(&data_);
  }

  MagmaCbc *ptr() const noexcept
  {
    return const_cast<MagmaCbc *>(reinterpret_cast<const MagmaCbc *>(&data_));
  }

  aligned_storage_t<MAGMA_CBC_SIZE, MAGMA_CBC_ALIGNMENT> data_;
};

/**
 * Потоковый шифр с длиной блока 128 бит в режиме гаммирования с обратной связью по шифртексту
 * с размером синхропосылки 16 байт. Определен в ГОСТ 34.13-2015 и ГОСТ 34.13-2018.
 *
 * \see #KuznyechikCfb
 */
class KuznyechikCfb_t
{
public:
  KuznyechikCfb_t(const KuznyechikCfb_t &) = default;
  KuznyechikCfb_t &operator=(const KuznyechikCfb_t &) = default;
  KuznyechikCfb_t(KuznyechikCfb_t &&) = default;
  KuznyechikCfb_t &operator=(KuznyechikCfb_t &&) = default;

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по шифртексту
   * с использованием ключа `key` и синхропосылки `nonce`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   * @param nonce — ненулевой указатель на массив длинной #KUZNYECHIK_CFB_NONCE_SIZE байт
   */
  explicit KuznyechikCfb_t(const uint8_t key[SYMMETRIC_KEY_SIZE], const uint8_t nonce[KUZNYECHIK_CFB_NONCE_SIZE]) noexcept
      : data_{}
  {
    kuznyechik_cfb_new(*ptr(), reinterpret_cast<const uint8_t *>(key), reinterpret_cast<const uint8_t *>(nonce));
  }

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по шифртексту
   * с использованием ключа `key` и синхропосылки `nonce`.
   *
   * @param key — ключ
   * @param nonce — синхропосылка
   */
  KuznyechikCfb_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key, const std::array<uint8_t, KUZNYECHIK_CFB_NONCE_SIZE> &nonce) noexcept
      : data_{}
  {
    kuznyechik_cfb_new(*ptr(), key.data(), nonce.data());
  }

  ~KuznyechikCfb_t() noexcept
  {
    kuznyechik_cfb_drop(*ptr());
  }

  /**
   * @brief Зашифровать сообщение.
   *
   * Зашифровывает сообщение на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   */
  void encrypt(uint8_t *buf, uintptr_t buf_len) noexcept
  {
    kuznyechik_cfb_encrypt(*ptr(), buf, buf_len);
  }

  /**
   * @brief Зашифровать сообщение.
   *
   * Зашифровывает сообщение на месте.
   *
   * @param buf — сообщение
   */
  void encrypt(std::vector<uint8_t> &buf) noexcept
  {
    encrypt(buf.data(), buf.size());
  }

  /**
   * @brief Расшифровать сообщение.
   *
   * Расшифровывает сообщение на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   */
  void decrypt(uint8_t *buf, uintptr_t buf_len) noexcept
  {
    kuznyechik_cfb_decrypt(*ptr(), buf, buf_len);
  }

  /**
   * @brief Расшифровать сообщение.
   *
   * Расшифровывает сообщение на месте.
   *
   * @param buf — сообщение
   */
  void decrypt(std::vector<uint8_t> &buf) noexcept
  {
    decrypt(buf.data(), buf.size());
  }

private:
  KuznyechikCfb *ptr() noexcept
  {
    return reinterpret_cast<KuznyechikCfb *>(&data_);
  }

  aligned_storage_t<KUZNYECHIK_CFB_SIZE, KUZNYECHIK_CFB_ALIGNMENT> data_;
};

/**
 * Потоковый шифр с длиной блока 64 бит в режиме гаммирования с обратной связью по шифртексту
 * с размером синхропосылки 8 байт. Определен в ГОСТ 34.13-2015 и ГОСТ 34.13-2018.
 *
 * \see #MagmaCfb
 */
class MagmaCfb_t
{
public:
  MagmaCfb_t(const MagmaCfb_t &) = default;
  MagmaCfb_t &operator=(const MagmaCfb_t &) = default;
  MagmaCfb_t(MagmaCfb_t &&) = default;
  MagmaCfb_t &operator=(MagmaCfb_t &&) = default;

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по шифртексту
   * с использованием ключа `key` и синхропосылки `nonce`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   * @param nonce — ненулевой указатель на массив длинной #MAGMA_CFB_NONCE_SIZE байт
   */
  explicit MagmaCfb_t(const uint8_t key[SYMMETRIC_KEY_SIZE], const uint8_t nonce[MAGMA_CFB_NONCE_SIZE]) noexcept
      : data_{}
  {
    magma_cfb_new(*ptr(), reinterpret_cast<const uint8_t *>(key), reinterpret_cast<const uint8_t *>(nonce));
  }

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по шифртексту
   * с использованием ключа `key` и синхропосылки `nonce`.
   *
   * @param key — ключ
   * @param nonce — синхропосылка
   */
  MagmaCfb_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key, const std::array<uint8_t, MAGMA_CFB_NONCE_SIZE> &nonce) noexcept
      : data_{}
  {
    magma_cfb_new(*ptr(), key.data(), nonce.data());
  }

  ~MagmaCfb_t() noexcept
  {
    magma_cfb_drop(*ptr());
  }

  /**
   * @brief Зашифровать сообщение.
   *
   * Зашифровывает сообщение на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   */
  void encrypt(uint8_t *buf, uintptr_t buf_len) noexcept
  {
    magma_cfb_encrypt(*ptr(), buf, buf_len);
  }

  /**
   * @brief Зашифровать сообщение.
   *
   * Зашифровывает сообщение на месте.
   *
   * @param buf — сообщение
   */
  void encrypt(std::vector<uint8_t> &buf) noexcept
  {
    encrypt(buf.data(), buf.size());
  }

  /**
   * @brief Расшифровать сообщение.
   *
   * Расшифровывает сообщение на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   */
  void decrypt(uint8_t *buf, uintptr_t buf_len) noexcept
  {
    magma_cfb_decrypt(*ptr(), buf, buf_len);
  }

  /**
   * @brief Расшифровать сообщение.
   *
   * Расшифровывает сообщение на месте.
   *
   * @param buf — сообщение
   */
  void decrypt(std::vector<uint8_t> &buf) noexcept
  {
    decrypt(buf.data(), buf.size());
  }

private:
  MagmaCfb *ptr() noexcept
  {
    return reinterpret_cast<MagmaCfb *>(&data_);
  }

  aligned_storage_t<MAGMA_CFB_SIZE, MAGMA_CFB_ALIGNMENT> data_;
};

/**
 * Блочный шифр с длиной блока 128 бит. Определён в ГОСТ Р 34.12-2015 (Кузнечик) и ГОСТ 34.12-2018.
 *
 * \see #Kuznyechik
 */
class Kuznyechik_t
{
public:
  Kuznyechik_t(const Kuznyechik_t &) = default;
  Kuznyechik_t &operator=(const Kuznyechik_t &) = default;
  Kuznyechik_t(Kuznyechik_t &&) = default;
  Kuznyechik_t &operator=(Kuznyechik_t &&) = default;

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт новый контекст блочного шифра.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределенному поведению.
   *
   * @param key — ненулевой указатель на ключ длиной #SYMMETRIC_KEY_SIZE байт.
   */
  explicit Kuznyechik_t(const uint8_t key[SYMMETRIC_KEY_SIZE]) noexcept
      : data_{}
  {
    kuznyechik_new(*ptr(), reinterpret_cast<const uint8_t *>(key));
  }

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт новый контекст блочного шифра.
   *
   * @param key — ключ
   */
  explicit Kuznyechik_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key) noexcept
      : data_{}
  {
    kuznyechik_new(*ptr(), key.data());
  }

  ~Kuznyechik_t() noexcept
  {
    kuznyechik_drop(*ptr());
  }

  /**
   * @brief Зашифровать блок данных.
   *
   * Шифрует блок на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределенному поведению.
   *
   * @param block — ненулевой указатель на блок длиной #KUZNYECHIK_BLOCK_SIZE байт.
   */
  void encrypt_block(uint8_t block[KUZNYECHIK_BLOCK_SIZE]) noexcept
  {
    kuznyechik_encrypt_block(*ptr(), block);
  }

  /**
   * @brief Расшифровать блок данных.
   *
   * Расшифровывает блок на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределенному поведению.
   *
   * @param block — ненулевой указатель на блок длиной #KUZNYECHIK_BLOCK_SIZE байт.
   */
  void decrypt_block(uint8_t block[KUZNYECHIK_BLOCK_SIZE]) noexcept
  {
    kuznyechik_decrypt_block(*ptr(), block);
  }

  /**
   * @brief Зашифровать массив блоков.
   *
   * Шифрует массив блоков на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределенному поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков, каждый блок длиной #KUZNYECHIK_BLOCK_SIZE байт.
   * @param len — длина массива блоков.
   */
  void encrypt_blocks(uint8_t blocks[KUZNYECHIK_BLOCK_SIZE], uintptr_t len) noexcept
  {
    kuznyechik_encrypt_blocks(*ptr(), blocks, len);
  }

  /**
   * @brief Расшифровать массив блоков.
   *
   * Расшифровывает массив блоков на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределенному поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков, каждый блок длиной #KUZNYECHIK_BLOCK_SIZE байт.
   * @param len — длина массива блоков.
   */
  void decrypt_blocks(uint8_t blocks[KUZNYECHIK_BLOCK_SIZE], uintptr_t len) noexcept
  {
    kuznyechik_decrypt_blocks(*ptr(), blocks, len);
  }

private:
  Kuznyechik *ptr() noexcept
  {
    return reinterpret_cast<Kuznyechik *>(&data_);
  }

  aligned_storage_t<KUZNYECHIK_SIZE, KUZNYECHIK_ALIGNMENT> data_;
};

/**
 * Блочный шифр с длиной блока 64 бит. Определён в ГОСТ Р 34.12-2015 (Магма) и ГОСТ 34.12-2018.
 *
 * \see #Magma
 */
class Magma_t
{
public:
  Magma_t(const Magma_t &) = default;
  Magma_t &operator=(const Magma_t &) = default;
  Magma_t(Magma_t &&) = default;
  Magma_t &operator=(Magma_t &&) = default;

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт новый контекст блочного шифра.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределенному поведению.
   *
   * @param key — ненулевой указатель на ключ длиной #SYMMETRIC_KEY_SIZE байт.
   */
  explicit Magma_t(const uint8_t key[SYMMETRIC_KEY_SIZE]) noexcept
      : data_{}
  {
    magma_new(*ptr(), reinterpret_cast<const uint8_t *>(key));
  }

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт новый контекст блочного шифра.
   *
   * @param key — ключ
   */
  explicit Magma_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key) noexcept
      : data_{}
  {
    magma_new(*ptr(), key.data());
  }

  ~Magma_t() noexcept
  {
    magma_drop(*ptr());
  }

  /**
   * @brief Зашифровать блок данных.
   *
   * Шифрует блок на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределенному поведению.
   *
   * @param block — ненулевой указатель на блок длиной #MAGMA_BLOCK_SIZE байт.
   */
  void encrypt_block(uint8_t block[MAGMA_BLOCK_SIZE]) noexcept
  {
    magma_encrypt_block(*ptr(), block);
  }

  /**
   * @brief Расшифровать блок данных.
   *
   * Расшифровывает блок на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределенному поведению.
   *
   * @param block — ненулевой указатель на блок длиной #MAGMA_BLOCK_SIZE байт.
   */
  void decrypt_block(uint8_t block[MAGMA_BLOCK_SIZE]) noexcept
  {
    magma_decrypt_block(*ptr(), block);
  }

  /**
   * @brief Зашифровать массив блоков.
   *
   * Шифрует массив блоков на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределенному поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков, каждый блок длиной #MAGMA_BLOCK_SIZE байт.
   * @param len — длина массива блоков.
   */
  void encrypt_blocks(uint8_t blocks[MAGMA_BLOCK_SIZE], uintptr_t len) noexcept
  {
    magma_encrypt_blocks(*ptr(), blocks, len);
  }

  /**
   * @brief Расшифровать массив блоков.
   *
   * Расшифровывает массив блоков на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределенному поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков, каждый блок длиной #MAGMA_BLOCK_SIZE байт.
   * @param len — длина массива блоков.
   */
  void decrypt_blocks(uint8_t blocks[MAGMA_BLOCK_SIZE], uintptr_t len) noexcept
  {
    magma_decrypt_blocks(*ptr(), blocks, len);
  }

private:
  Magma *ptr() noexcept
  {
    return reinterpret_cast<Magma *>(&data_);
  }

  aligned_storage_t<MAGMA_SIZE, MAGMA_ALIGNMENT> data_;
};

/**
 * Потоковый шифр с длиной блока 128 бит в режиме гаммирования. Определен в ГОСТ 34.13-2015 и ГОСТ 34.13-2018.
 *
 * \see #KuznyechikCtr
 */
class KuznyechikCtr_t
{
public:
  KuznyechikCtr_t(const KuznyechikCtr_t &) = default;
  KuznyechikCtr_t &operator=(const KuznyechikCtr_t &) = default;
  KuznyechikCtr_t(KuznyechikCtr_t &&) = default;
  KuznyechikCtr_t &operator=(KuznyechikCtr_t &&) = default;

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с использованием ключа `key`
   * и синхропосылки `nonce`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   * @param nonce — ненулевой указатель на массив длинной #KUZNYECHIK_CTR_NONCE_SIZE байт
   */
  explicit KuznyechikCtr_t(const uint8_t key[SYMMETRIC_KEY_SIZE], const uint8_t nonce[KUZNYECHIK_CTR_NONCE_SIZE]) noexcept
      : data_{}
  {
    kuznyechik_ctr_new(*ptr(), reinterpret_cast<const uint8_t *>(key), reinterpret_cast<const uint8_t *>(nonce));
  }

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с использованием ключа `key`
   * и синхропосылки `nonce`.
   *
   * @param key — ключ
   * @param nonce — синхропосылка
   */
  KuznyechikCtr_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key, const std::array<uint8_t, KUZNYECHIK_CTR_NONCE_SIZE> &nonce) noexcept
      : data_{}
  {
    kuznyechik_ctr_new(*ptr(), key.data(), nonce.data());
  }

  ~KuznyechikCtr_t() noexcept
  {
    kuznyechik_ctr_drop(*ptr());
  }

  /**
   * @brief Гаммировать сообщение ключевым потоком шифра.
   *
   * Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
   * шифра. Возвращает ошибку в случае, когда достигнут конец ключевого потока.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   *
   * @returns 0, в случае успеха, -1, в случае ошибки
   */
  intptr_t apply_keystream(uint8_t *buf, uintptr_t buf_len) noexcept
  {
    return kuznyechik_ctr_apply_keystream(*ptr(), buf, buf_len);
  }

  /**
   * @brief Гаммировать сообщение ключевым потоком шифра.
   *
   * Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
   * шифра. Возвращает ошибку в случае, когда достигнут конец ключевого потока.
   *
   * @param buf — сообщение
   *
   * @returns 0, в случае успеха, -1, в случае ошибки
   */
  intptr_t apply_keystream(std::vector<uint8_t> &buf) noexcept
  {
    return apply_keystream(buf.data(), buf.size());
  }

  /**
   * @brief Установить позицию ключевого потока шифра.
   *
   * Устанавливает текущую позицию ключевого потока шифра. Возвращает ошибку в случае,
   * когда требуемая позиция превышает длину ключевого потока.
   *
   * @param pos — новая позиция ключевого потока
   *
   * @returns 0, в случае успеха, -1, в случае ошибки
   */
  intptr_t seek(uint64_t pos) noexcept
  {
    return kuznyechik_ctr_seek(*ptr(), pos);
  }

  /**
   * @brief Получить позицию ключевого потока шифра.
   *
   * Возвращает текущую позицию ключевого потока шифра.
   *
   * @returns Текущая позиция ключевого потока
   */
  uint64_t get_pos() const noexcept
  {
    return kuznyechik_ctr_get_pos(*ptr());
  }

private:
  KuznyechikCtr *ptr() noexcept
  {
    return reinterpret_cast<KuznyechikCtr *>(&data_);
  }

  KuznyechikCtr *ptr() const noexcept
  {
    return const_cast<KuznyechikCtr *>(reinterpret_cast<const KuznyechikCtr *>(&data_));
  }

  aligned_storage_t<KUZNYECHIK_CTR_SIZE, KUZNYECHIK_CTR_ALIGNMENT> data_;
};

/**
 * Потокововый шифр с длиной блока 64 бит в режиме гаммирования. Определен в ГОСТ 34.13-2015 и ГОСТ 34.13-2018.
 *
 * \see #MagmaCtr
 */
class MagmaCtr_t
{
public:
  MagmaCtr_t(const MagmaCtr_t &) = default;
  MagmaCtr_t &operator=(const MagmaCtr_t &) = default;
  MagmaCtr_t(MagmaCtr_t &&) = default;
  MagmaCtr_t &operator=(MagmaCtr_t &&) = default;

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с использованием ключа `key`
   * и синхропосылки `nonce`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   * @param nonce — ненулевой указатель на массив длинной #MAGMA_CTR_NONCE_SIZE байт
   */
  explicit MagmaCtr_t(const uint8_t key[SYMMETRIC_KEY_SIZE], const uint8_t nonce[MAGMA_CTR_NONCE_SIZE]) noexcept
      : data_{}
  {
    magma_ctr_new(*ptr(), reinterpret_cast<const uint8_t *>(key), reinterpret_cast<const uint8_t *>(nonce));
  }

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с использованием ключа `key`
   * и синхропосылки `nonce`.
   *
   * @param key — ключ
   * @param nonce — синхропосылка
   */
  MagmaCtr_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key, const std::array<uint8_t, MAGMA_CTR_NONCE_SIZE> &nonce) noexcept
      : data_{}
  {
    magma_ctr_new(*ptr(), key.data(), nonce.data());
  }

  ~MagmaCtr_t() noexcept
  {
    magma_ctr_drop(*ptr());
  }

  /**
   * @brief Гаммировать сообщение ключевым потоком шифра.
   *
   * Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
   * шифра. Возвращает ошибку в случае, когда достигнут конец ключевого потока.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   *
   * @returns 0, в случае успеха, -1, в случае ошибки
   */
  intptr_t apply_keystream(uint8_t *buf, uintptr_t buf_len) noexcept
  {
    return magma_ctr_apply_keystream(*ptr(), buf, buf_len);
  }

  /**
   * @brief Гаммировать сообщение ключевым потоком шифра.
   *
   * Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
   * шифра. Возвращает ошибку в случае, когда достигнут конец ключевого потока.
   *
   * @param buf — сообщение
   *
   * @returns 0, в случае успеха, -1, в случае ошибки
   */
  intptr_t apply_keystream(std::vector<uint8_t> &buf) noexcept
  {
    return apply_keystream(buf.data(), buf.size());
  }

  /**
   * @brief Установить позицию ключевого потока шифра.
   *
   * Устанавливает текущую позицию ключевого потока шифра. Возвращает ошибку в случае,
   * когда требуемая позиция превышает длину ключевого потока.
   *
   * @param pos — новая позиция ключевого потока
   *
   * @returns 0, в случае успеха, -1, в случае ошибки
   */
  intptr_t seek(uint64_t pos) noexcept
  {
    return magma_ctr_seek(*ptr(), pos);
  }

  /**
   * @brief Получить позицию ключевого потока шифра.
   *
   * Возвращает текущую позицию ключевого потока шифра.
   *
   * @returns Текущая позиция ключевого потока
   */
  uint64_t get_pos() const noexcept
  {
    return magma_ctr_get_pos(*ptr());
  }

private:
  MagmaCtr *ptr() noexcept
  {
    return reinterpret_cast<MagmaCtr *>(&data_);
  }

  MagmaCtr *ptr() const noexcept
  {
    return const_cast<MagmaCtr *>(reinterpret_cast<const MagmaCtr *>(&data_));
  }

  aligned_storage_t<MAGMA_CTR_SIZE, MAGMA_CTR_ALIGNMENT> data_;
};

/**
 * Блочный шифр с длиной блока 128 бит в режиме простой замены. Определен в ГОСТ 34.13-2015 и ГОСТ 34.13-2018.
 *
 * \see #KuznyechikEcb
 */
class KuznyechikEcb_t
{
public:
  KuznyechikEcb_t(const KuznyechikEcb_t &) = default;
  KuznyechikEcb_t &operator=(const KuznyechikEcb_t &) = default;
  KuznyechikEcb_t(KuznyechikEcb_t &&) = default;
  KuznyechikEcb_t &operator=(KuznyechikEcb_t &&) = default;

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
   * ключа `key`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   */
  explicit KuznyechikEcb_t(const uint8_t key[SYMMETRIC_KEY_SIZE]) noexcept
      : data_{}
  {
    kuznyechik_ecb_new(*ptr(), reinterpret_cast<const uint8_t *>(key));
  }

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
   * ключа `key`.
   *
   * @param key — ключ
   */
  explicit KuznyechikEcb_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key) noexcept
      : data_{}
  {
    kuznyechik_ecb_new(*ptr(), key.data());
  }

  ~KuznyechikEcb_t() noexcept
  {
    kuznyechik_ecb_drop(*ptr());
  }

  /**
   * @brief Зашифровать массив блоков.
   *
   * Шифрует блоки данных на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков. Блок представляет собой последовательность байт
   * длиной #KUZNYECHIK_BLOCK_SIZE.
   * @param len — длина массива блоков (в блоках)
   */
  void encrypt_blocks(uint8_t (*blocks)[KUZNYECHIK_BLOCK_SIZE], uintptr_t len) noexcept
  {
    kuznyechik_ecb_encrypt_blocks(*ptr(), reinterpret_cast<uint8_t *>(blocks), len);
  }

  /**
   * @brief Расшифровать массив блоков.
   *
   * Расшифровывает блоки данных на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков. Блок представляет собой последовательность байт
   * длиной #KUZNYECHIK_BLOCK_SIZE.
   * @param len — длина массива блоков (в блоках)
   */
  void decrypt_blocks(uint8_t (*blocks)[KUZNYECHIK_BLOCK_SIZE], uintptr_t len) noexcept
  {
    kuznyechik_ecb_decrypt_blocks(*ptr(), reinterpret_cast<uint8_t *>(blocks), len);
  }

  /**
   * @brief Зашифровать сообщение.
   *
   * Зашифровывает сообщение на месте.
   *
   * Байты `0..pos` буфера `buf` используются как сообщение, когда же байты `pos..buf_len`
   * используются как зарезервированное место для дополнения.
   *
   * Возвращает ошибку в случае, когда место для дополнения недостаточно велико.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   * @param pos — индекс байта, с которого начинается дополнение в буфере `buf`
   *
   * @returns Длину шифротекста, в случае успеха или -1, в случае ошибки
   */
  intptr_t encrypt_pad(uint8_t *buf, uintptr_t buf_len, uintptr_t pos) const noexcept
  {
    auto copy = *this;
    return kuznyechik_ecb_encrypt_pad(*copy.ptr(), buf, buf_len, pos);
  }

  /**
   * @brief Расшифровать сообщение.
   *
   * Расшифровывает сообщение на месте.
   *
   * Возвращает ошибку в случае, когда длина буфера не делится на #KUZNYECHIK_BLOCK_SIZE или
   * расшифрованное сообщение имеет некорректное дополнение.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   *
   * @returns Длину сообщения, в случае успеха или -1, в случае ошибки
   */
  intptr_t decrypt_pad(uint8_t *buf, uintptr_t buf_len) const noexcept
  {
    auto copy = *this;
    return kuznyechik_ecb_decrypt_pad(*copy.ptr(), buf, buf_len);
  }

private:
  KuznyechikEcb *ptr() noexcept
  {
    return reinterpret_cast<KuznyechikEcb *>(&data_);
  }

  KuznyechikEcb *ptr() const noexcept
  {
    return const_cast<KuznyechikEcb *>(reinterpret_cast<const KuznyechikEcb *>(&data_));
  }

  aligned_storage_t<KUZNYECHIK_ECB_SIZE, KUZNYECHIK_ECB_ALIGNMENT> data_;
};

/**
 * Блочный шифр с длиной блока 64 бит в режиме простой замены. Определен в ГОСТ 34.13-2015 и ГОСТ 34.13-2018.
 *
 * \see #MagmaEcb
 */
class MagmaEcb_t
{
public:
  MagmaEcb_t(const MagmaEcb_t &) = default;
  MagmaEcb_t &operator=(const MagmaEcb_t &) = default;
  MagmaEcb_t(MagmaEcb_t &&) = default;
  MagmaEcb_t &operator=(MagmaEcb_t &&) = default;

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
   * ключа `key`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   */
  explicit MagmaEcb_t(const uint8_t key[SYMMETRIC_KEY_SIZE]) noexcept
      : data_{}
  {
    magma_ecb_new(*ptr(), reinterpret_cast<const uint8_t *>(key));
  }

  /**
   * @brief Создать контекст блочного шифра.
   *
   * Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
   * ключа `key`.
   *
   * @param key — ключ
   */
  explicit MagmaEcb_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key) noexcept
      : data_{}
  {
    magma_ecb_new(*ptr(), key.data());
  }

  ~MagmaEcb_t() noexcept
  {
    magma_ecb_drop(*ptr());
  }

  /**
   * @brief Зашифровать массив блоков.
   *
   * Шифрует блоки данных на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков. Блок представляет собой последовательность
   * байт длиной #MAGMA_BLOCK_SIZE.
   * @param len — длина массива блоков (в блоках)
   */
  void encrypt_blocks(uint8_t (*blocks)[MAGMA_BLOCK_SIZE], uintptr_t len) noexcept
  {
    magma_ecb_encrypt_blocks(*ptr(), reinterpret_cast<uint8_t *>(blocks), len);
  }

  /**
   * @brief Расшифровать массив блоков.
   *
   * Расшифровывает блоки данных на месте.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param blocks — ненулевой указатель на массив блоков. Блок представляет собой последовательность байт
   * длиной #MAGMA_BLOCK_SIZE.
   * @param len — длина массива блоков (в блоках)
   */
  void decrypt_blocks(uint8_t (*blocks)[MAGMA_BLOCK_SIZE], uintptr_t len) noexcept
  {
    magma_ecb_decrypt_blocks(*ptr(), reinterpret_cast<uint8_t *>(blocks), len);
  }

  /**
   * @brief Зашифровать сообщение.
   *
   * Зашифровывает сообщение на месте.
   *
   * Байты `0..pos` буфера `buf` используются как сообщение, когда же байты `pos..buf_len`
   * используются как зарезервированное место для дополнения.
   *
   * Возвращает ошибку в случае, когда место для дополнения недостаточно велико.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   * @param pos — индекс байта, с которого начинается дополнение в буфере `buf`
   *
   * @returns Длину шифротекста, в случае успеха или -1, в случае ошибки
   */
  intptr_t encrypt_pad(uint8_t *buf, uintptr_t buf_len, uintptr_t pos) const noexcept
  {
    auto copy = *this;
    return magma_ecb_encrypt_pad(*copy.ptr(), buf, buf_len, pos);
  }

  /**
   * @brief Расшифровать сообщение.
   *
   * Расшифровывает сообщение на месте.
   *
   * Возвращает ошибку в случае, когда длина буфера не делится на #MAGMA_BLOCK_SIZE или
   * расшифрованное сообщение имеет некорректное дополнение.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   *
   * @returns Длину сообщения, в случае успеха или -1, в случае ошибки
   */
  intptr_t decrypt_pad(uint8_t *buf, uintptr_t buf_len) const noexcept
  {
    auto copy = *this;
    return magma_ecb_decrypt_pad(*copy.ptr(), buf, buf_len);
  }

private:
  MagmaEcb *ptr() noexcept
  {
    return reinterpret_cast<MagmaEcb *>(&data_);
  }

  MagmaEcb *ptr() const noexcept
  {
    return const_cast<MagmaEcb *>(reinterpret_cast<const MagmaEcb *>(&data_));
  }

  aligned_storage_t<MAGMA_ECB_SIZE, MAGMA_ECB_ALIGNMENT> data_;
};

/**
 * Функция хэширования с длиной хэш-кода 256 бит. Определена в ГОСТ Р 34.11-2012 (Стрибог) и ГОСТ 34.11-2018.
 */
class Streebog256_t
{
public:
  Streebog256_t(const Streebog256_t &) = default;
  Streebog256_t &operator=(const Streebog256_t &) = default;
  Streebog256_t(Streebog256_t &&) = default;
  Streebog256_t &operator=(Streebog256_t &&) = default;

  /**
   * @brief Создать состояние функции хэширования.
   *
   * Создаёт новое состояние функции хэширования.
   */
  Streebog256_t() noexcept
      : data_{}
  {
    streebog256_new(*ptr());
  }

  ~Streebog256_t() noexcept
  {
    streebog256_drop(*ptr());
  }

  /**
   * @brief Хэшировать сообщение.
   *
   * Хэширует массив данных `msg` и копирует результирующий хэш в `res`.
   *
   * \warning Передача нулевых или некорректных указателей приводит к неопределённому поведению.
   *
   * @param msg — ненулевой указатель на массив байт
   * @param len — длина сообщения в байтах
   * @param res — массив длиной не менее чем #STREEBOG256_RESULT_SIZE байт
   */
  static void digest(const uint8_t *msg, uintptr_t len, uint8_t res[STREEBOG256_RESULT_SIZE]) noexcept
  {
    streebog256_digest(msg, len, reinterpret_cast<uint8_t *>(res));
  }

  /**
   * @brief Хэшировать сообщение.
   *
   * Хэширует массив данных `msg` и возвращает хэш.
   *
   * \warning Передача нулевых или некорректных указателей приводит к неопределённому поведению.
   *
   * @param msg — ненулевой указатель на массив байт
   * @param len — длина сообщения в байтах
   *
   * @returns хэш как массив
   */
  static std::array<uint8_t, STREEBOG256_RESULT_SIZE> digest(const uint8_t *msg, uintptr_t len) noexcept
  {
    std::array<uint8_t, STREEBOG256_RESULT_SIZE> result {};
    digest(msg, len, result.data());
    return result;
  }

  /**
   * @brief Обновить состояние функции хэширования.
   *
   * Скармливает сообщение функции хэширования, обновляя её состояние.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param msg — ненулевой указатель на массив байт
   * @param len — длина сообщения в байтах
   */
  void update(const uint8_t *msg, uintptr_t len) noexcept
  {
    streebog256_update(*ptr(), msg, len);
  }

  /**
   * @brief Обновить состояние функции хэширования.
   *
   * Скармливает сообщение функции хэширования, обновляя её состояние.
   *
   * @param message — сообщение
   */
  void update(const std::vector<uint8_t> &message) noexcept
  {
    streebog256_update(*ptr(), message.data(), message.size());
  }

  /**
   * @brief Финализировать состояние функции хэширования.
   *
   * Финализирует состояние функции хэширования и копирует результирующий хэш в `res`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param res — массив длиной не менее чем #STREEBOG256_RESULT_SIZE байт
   */
  void finalize(uint8_t res[STREEBOG256_RESULT_SIZE]) const noexcept
  {
    auto copy = *this;
    streebog256_finalize(*copy.ptr(), reinterpret_cast<uint8_t *>(res));
  }

  /**
   * @brief Финализировать состояние функции хэширования.
   *
   * Финализирует состояние функции хэширования и возвращает хэш.
   *
   * @returns финализированный массив
   */
  std::array<uint8_t, STREEBOG256_RESULT_SIZE> finalize() const noexcept
  {
    std::array<uint8_t, STREEBOG256_RESULT_SIZE> result {};
    auto copy = *this;
    streebog256_finalize(*copy.ptr(), result.data());
    return result;
  }

  /**
   * @brief Финализировать состояние функции хэширования и заменить его новым.
   *
   * Финализирует состояние функции хэширования и копирует результирующий хэш в `res`.
   * Заменяет существующий состояние на новый.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param res — массив длиной не менее чем #STREEBOG256_RESULT_SIZE байт
   */
  void finalize_reset(uint8_t res[STREEBOG256_RESULT_SIZE]) noexcept
  {
    streebog256_finalize_reset(*ptr(), reinterpret_cast<uint8_t *>(res));
  }

  /**
   * @brief Финализировать состояние функции хэширования и заменить его новым.
   *
   * Финализирует состояние функции хэширования и возвращает хэш.
   * Заменяет существующий состояние на новый.
   *
   * @returns — финализированный массив
   */
  std::array<uint8_t, STREEBOG256_RESULT_SIZE> finalize_reset() noexcept
  {
    std::array<uint8_t, STREEBOG256_RESULT_SIZE> result {};
    streebog256_finalize_reset(*ptr(), result.data());
    return result;
  }

  /**
   * @brief Сбросить состояние функции хэширования.
   *
   * Заменяет состояние функции хэширования новым состоянием.
   */
  void reset() noexcept
  {
    streebog256_reset(*ptr());
  }

private:
  Streebog256 *ptr() const noexcept
  {
    return const_cast<Streebog256 *>(reinterpret_cast<const Streebog256 *>(&data_));
  }

  Streebog256 *ptr() noexcept
  {
    return reinterpret_cast<Streebog256 *>(&data_);
  }

  aligned_storage_t<STREEBOG256_SIZE, STREEBOG256_ALIGNMENT> data_;
};

/**
 * Функция хэширования с длиной хэш-кода 512 бит. Определена в ГОСТ Р 34.11-2012 (Стрибог) и ГОСТ 34.11-2018.
 */
class Streebog512_t
{
public:
  Streebog512_t(const Streebog512_t &) = default;
  Streebog512_t &operator=(const Streebog512_t &) = default;
  Streebog512_t(Streebog512_t &&) = default;
  Streebog512_t &operator=(Streebog512_t &&) = default;

  /**
   * @brief Создать состояние функции хэширования.
   *
   * Создаёт новое состояние функции хэширования.
   */
  Streebog512_t() noexcept
      : data_{}
  {
    streebog512_new(*ptr());
  }

  ~Streebog512_t() noexcept
  {
    streebog512_drop(*ptr());
  }

  /**
   * @brief Хэшировать сообщение.
   *
   * Хэширует массив данных `msg` и копирует результирующий хэш в `res`.
   *
   * \warning Передача нулевых или некорректных указателей приводит к неопределённому поведению.
   *
   * @param msg — ненулевой указатель на массив байт
   * @param len — длина сообщения в байтах
   * @param res — массив длиной не менее чем #STREEBOG512_RESULT_SIZE байт
   */
  static void digest(const uint8_t *msg, uintptr_t len, uint8_t res[STREEBOG512_RESULT_SIZE]) noexcept
  {
    streebog512_digest(msg, len, reinterpret_cast<uint8_t *>(res));
  }

  /**
   * @brief Хэшировать сообщение.
   *
   * Хэширует массив данных `msg` и возвращает хэш.
   *
   * \warning Передача нулевых или некорректных указателей приводит к неопределённому поведению.
   *
   * @param msg — ненулевой указатель на массив байт
   * @param len — длина сообщения в байтах
   *
   * @returns хэш как массив
   */
  static std::array<uint8_t, STREEBOG512_RESULT_SIZE> digest(const uint8_t *msg, uintptr_t len) noexcept
  {
    std::array<uint8_t, STREEBOG512_RESULT_SIZE> result {};
    digest(msg, len, result.data());
    return result;
  }

  /**
   * @brief Обновить состояние функции хэширования.
   *
   * Скармливает сообщение функции хэширования, обновляя её состояние.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param msg — ненулевой указатель на массив байт
   * @param len — длина сообщения в байтах
   */
  void update(const uint8_t *msg, uintptr_t len) noexcept
  {
    streebog512_update(*ptr(), msg, len);
  }

  /**
   * @brief Обновить состояние функции хэширования.
   *
   * Скармливает сообщение функции хэширования, обновляя её состояние.
   *
   * @param message — сообщение
   */
  void update(const std::vector<uint8_t> &message) noexcept
  {
    streebog512_update(*ptr(), message.data(), message.size());
  }

  /**
   * @brief Финализировать состояние функции хэширования.
   *
   * Финализирует состояние функции хэширования и копирует результирующий хэш в `res`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param res — массив длиной не менее чем #STREEBOG512_RESULT_SIZE байт
   */
  void finalize(uint8_t res[STREEBOG512_RESULT_SIZE]) const noexcept
  {
    auto copy = *this;
    streebog512_finalize(*copy.ptr(), reinterpret_cast<uint8_t *>(res));
  }

  /**
   * @brief Финализировать состояние функции хэширования.
   *
   * Финализирует состояние функции хэширования и возвращает хэш.
   *
   * @returns финализированный массив
   */
  std::array<uint8_t, STREEBOG512_RESULT_SIZE> finalize() const noexcept
  {
    std::array<uint8_t, STREEBOG512_RESULT_SIZE> result {};
    auto copy = *this;
    streebog512_finalize(*copy.ptr(), result.data());
    return result;
  }

  /**
   * @brief Финализировать состояние функции хэширования и заменить его новым.
   *
   * Финализирует состояние функции хэширования и копирует результирующий хэш в `res`.
   * Заменяет существующий состояние на новый.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param res — массив длиной не менее чем #STREEBOG512_RESULT_SIZE байт
   */
  void finalize_reset(uint8_t res[STREEBOG512_RESULT_SIZE]) noexcept
  {
    streebog512_finalize_reset(*ptr(), reinterpret_cast<uint8_t *>(res));
  }

  /**
   * @brief Финализировать состояние функции хэширования и заменить его новым.
   *
   * Финализирует состояние функции хэширования и возвращает хэш.
   * Заменяет существующий состояние на новый.
   *
   * @returns — финализированный массив
   */
  std::array<uint8_t, STREEBOG512_RESULT_SIZE> finalize_reset() noexcept
  {
    std::array<uint8_t, STREEBOG512_RESULT_SIZE> result {};
    streebog512_finalize_reset(*ptr(), result.data());
    return result;
  }

  /**
   * @brief Сбросить состояние функции хэширования.
   *
   * Заменяет состояние функции хэширования новым состоянием.
   */
  void reset() noexcept
  {
    streebog512_reset(*ptr());
  }

private:
  Streebog512 *ptr() const noexcept
  {
    return const_cast<Streebog512 *>(reinterpret_cast<const Streebog512 *>(&data_));
  }

  Streebog512 *ptr() noexcept
  {
    return reinterpret_cast<Streebog512 *>(&data_);
  }

  aligned_storage_t<STREEBOG512_SIZE, STREEBOG512_ALIGNMENT> data_;
};

/**
 * Режим выработки имитовставки с использованием шифра с длиной блока 128 бит. Определен в ГОСТ 34.13-2015 и ГОСТ 34.13-2018.
 *
 * |see #KuznyechikMac
 */
class KuznyechikMac_t
{
public:
  KuznyechikMac_t(const KuznyechikMac_t &) = default;
  KuznyechikMac_t &operator=(const KuznyechikMac_t &) = default;
  KuznyechikMac_t(KuznyechikMac_t &&) = default;
  KuznyechikMac_t &operator=(KuznyechikMac_t &&) = default;

  /**
   * @brief Создать состояние режима выработки имитовставки.
   *
   * Создаёт новое состояние режима выработки имитовставки с использованием ключа `key`.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   */
  explicit KuznyechikMac_t(const uint8_t key[SYMMETRIC_KEY_SIZE]) noexcept
      : data_{}
  {
    kuznyechik_mac_new(*ptr(), key);
  }

  /**
   * @brief Создать состояние режима выработки имитовставки.
   *
   * Создаёт новое состояние режима выработки имитовставки с использованием ключа `key`.
   *
   * @param key — ключ
   */
  explicit KuznyechikMac_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key) noexcept
      : data_{}
  {
    kuznyechik_mac_new(*ptr(), key.data());
  }

  ~KuznyechikMac_t() noexcept
  {
    kuznyechik_mac_drop(*ptr());
  }

  /**
   * @brief Обновить состояние режиму выработки имитовставки.
   *
   * Скармливает сообщение режиму выработки имитовставки, обновляя его состояние.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param msg — ненулевой указатель на массив байт
   * @param len — длина сообщения в байтах
   */
  void update(const uint8_t *msg, uintptr_t len) noexcept
  {
    kuznyechik_mac_update(*ptr(), msg, len);
  }

  /**
   * @brief Обновить состояние режиму выработки имитовставки.
   *
   * Скармливает сообщение режиму выработки имитовставки, обновляя его состояние.
   *
   * @param message — сообщение
   */
  void update(const std::vector<uint8_t> &message) noexcept
  {
    kuznyechik_mac_update(*ptr(), message.data(), message.size());
  }

  /**
   * @brief Финализировать состояние режима выработки имитовставки.
   *
   * Финализирует состояние режима выработки имитовставки и копирует результирующую имитовставку
   * в массив `res`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param res — массив длиной не менее чем #KUZNYECHIK_MAC_RESULT_SIZE байт
   */
  void finalize(uint8_t res[KUZNYECHIK_MAC_RESULT_SIZE]) const noexcept
  {
    auto copy = *this;
    kuznyechik_mac_finalize(*copy.ptr(), reinterpret_cast<uint8_t *>(res));
  }

  /**
   * @brief Финализировать состояние режима выработки имитовставки.
   *
   * Финализирует состояние режима выработки имитовставки и возвращает результирующую имитовставку.
   *
   * @returns — имитовставка
   */
  std::array<uint8_t, KUZNYECHIK_MAC_RESULT_SIZE> finalize() const noexcept
  {
    std::array<uint8_t, KUZNYECHIK_MAC_RESULT_SIZE> result {};
    auto copy = *this;
    kuznyechik_mac_finalize(*copy.ptr(), result.data());
    return result;
  }

  /**
   * @brief Финализировать состояние режима выработки имитовставки и заменить его новым.
   *
   * Финализирует состояние режима выработки имитовставки и копирует результирующую имитовставку
   * в массив `res`. Заменяет существующее состояние на новое.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param res — массив длиной не менее чем #KUZNYECHIK_MAC_RESULT_SIZE байт
   */
  void finalize_reset(uint8_t res[KUZNYECHIK_MAC_RESULT_SIZE]) noexcept
  {
    kuznyechik_mac_finalize_reset(*ptr(), reinterpret_cast<uint8_t *>(res));
  }

  /**
   * @brief Финализировать состояние режима выработки имитовставки и заменить его новым.
   *
   * Финализирует состояние режима выработки имитовставки и возвращает результирующую имитовставку.
   * Заменяет существующий состояние на новое.
   *
   * @returns — имитовставка
   */
  std::array<uint8_t, KUZNYECHIK_MAC_RESULT_SIZE> finalize_reset() noexcept
  {
    std::array<uint8_t, KUZNYECHIK_MAC_RESULT_SIZE> result {};
    kuznyechik_mac_finalize_reset(*ptr(), result.data());
    return result;
  }

  /**
   * @brief Проверить целостность данных.
   *
   * Проверяет соответствие данных, скормленных режиму выработки имитовставки, данной имитовставке.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param code — массив байт длиной #KUZNYECHIK_MAC_RESULT_SIZE
   *
   * @returns true, в случае соответствия, false в случае несоответствия
   */
  bool verify(const uint8_t code[KUZNYECHIK_MAC_RESULT_SIZE]) const noexcept
  {
    return kuznyechik_mac_verify(*ptr(), reinterpret_cast<const uint8_t *>(code));
  }

  /**
   * @brief Проверить целостность данных.
   *
   * Проверяет соответствие данных, скормленных режиму выработки имитовставки, данной имитовставке.
   *
   * @param code — имитовставка
   *
   * @returns true, в случае соответствия, false в случае несоответствия
   */
  bool verify(const std::array<uint8_t, KUZNYECHIK_MAC_RESULT_SIZE> &code) const noexcept
  {
    return kuznyechik_mac_verify(*ptr(), code.data());
  }

  /**
   * @brief Сбросить состояние режима выработки имитовставки.
   *
   * Заменяет состояние режима выработки имитовставки новым состоянием.
   */
  void reset() noexcept
  {
    kuznyechik_mac_reset(*ptr());
  }

private:
  KuznyechikMac *ptr() const noexcept
  {
    return const_cast<KuznyechikMac *>(reinterpret_cast<const KuznyechikMac *>(&data_));
  }

  KuznyechikMac *ptr() noexcept
  {
    return reinterpret_cast<KuznyechikMac *>(&data_);
  }

  aligned_storage_t<KUZNYECHIK_MAC_SIZE, KUZNYECHIK_MAC_ALIGNMENT> data_;
};

/**
 * Режим выработки имитовставки с использованием шифра с длиной блока 64 бит. Определен в ГОСТ 34.13-2015 и ГОСТ 34.13-2018.
 *
 * \see #MagmaMac
 */
class MagmaMac_t
{
public:
  MagmaMac_t(const MagmaMac_t &) = default;
  MagmaMac_t &operator=(const MagmaMac_t &) = default;
  MagmaMac_t(MagmaMac_t &&) = default;
  MagmaMac_t &operator=(MagmaMac_t &&) = default;

  /**
   * @brief Создать состояние режима выработки имитовставки.
   *
   * Создаёт новое состояние режима выработки имитовставки с использованием ключа `key`.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   */
  explicit MagmaMac_t(const uint8_t key[SYMMETRIC_KEY_SIZE]) noexcept
      : data_{}
  {
    magma_mac_new(*ptr(), key);
  }

  /**
   * @brief Создать состояние режима выработки имитовставки.
   *
   * Создаёт новое состояние режима выработки имитовставки с использованием ключа `key`.
   *
   * @param key — ключ
   */
  explicit MagmaMac_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key) noexcept
      : data_{}
  {
    magma_mac_new(*ptr(), key.data());
  }

  ~MagmaMac_t() noexcept
  {
    magma_mac_drop(*ptr());
  }

  /**
   * @brief Обновить состояние режиму выработки имитовставки.
   *
   * Скармливает сообщение режиму выработки имитовставки, обновляя его состояние.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param msg — ненулевой указатель на массив байт
   * @param len — длина сообщения в байтах
   */
  void update(const uint8_t *msg, uintptr_t len) noexcept
  {
    magma_mac_update(*ptr(), msg, len);
  }

  /**
   * @brief Обновить состояние режиму выработки имитовставки.
   *
   * Скармливает сообщение режиму выработки имитовставки, обновляя его состояние.
   *
   * @param message — сообщение
   */
  void update(const std::vector<uint8_t> &message) noexcept
  {
    magma_mac_update(*ptr(), message.data(), message.size());
  }

  /**
   * @brief Финализировать состояние режима выработки имитовставки.
   *
   * Финализирует состояние режима выработки имитовставки и копирует результирующую имитовставку
   * в массив `res`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param res — массив длиной не менее чем #MAGMA_MAC_RESULT_SIZE байт
   */
  void finalize(uint8_t res[MAGMA_MAC_RESULT_SIZE]) const noexcept
  {
    auto copy = *this;
    magma_mac_finalize(*copy.ptr(), reinterpret_cast<uint8_t *>(res));
  }

  /**
   * @brief Финализировать состояние режима выработки имитовставки.
   *
   * Финализирует состояние режима выработки имитовставки и возвращает результирующую имитовставку.
   *
   * @returns — имитовставка
   */
  std::array<uint8_t, MAGMA_MAC_RESULT_SIZE> finalize() const noexcept
  {
    std::array<uint8_t, MAGMA_MAC_RESULT_SIZE> result {};
    auto copy = *this;
    magma_mac_finalize(*copy.ptr(), result.data());
    return result;
  }

  /**
   * @brief Финализировать состояние режима выработки имитовставки и заменить его новым.
   *
   * Финализирует состояние режима выработки имитовставки и копирует результирующую имитовставку
   * в массив `res`. Заменяет существующее состояние на новое.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param res — массив длиной не менее чем #MAGMA_MAC_RESULT_SIZE байт
   */
  void finalize_reset(uint8_t res[MAGMA_MAC_RESULT_SIZE]) noexcept
  {
    magma_mac_finalize_reset(*ptr(), reinterpret_cast<uint8_t *>(res));
  }

  /**
   * @brief Финализировать состояние режима выработки имитовставки и заменить его новым.
   *
   * Финализирует состояние режима выработки имитовставки и возвращает результирующую имитовставку.
   * Заменяет существующий состояние на новое.
   *
   * @returns — имитовставка
   */
  std::array<uint8_t, MAGMA_MAC_RESULT_SIZE> finalize_reset() noexcept
  {
    std::array<uint8_t, MAGMA_MAC_RESULT_SIZE> result {};
    magma_mac_finalize_reset(*ptr(), result.data());
    return result;
  }

  /**
   * @brief Проверить целостность данных.
   *
   * Проверяет соответствие данных, скормленных режиму выработки имитовставки, данной имитовставке.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param code — массив байт длиной #MAGMA_MAC_RESULT_SIZE
   *
   * @returns true, в случае соответствия, false в случае несоответствия
   */
  bool verify(const uint8_t code[MAGMA_MAC_RESULT_SIZE]) const noexcept
  {
    return magma_mac_verify(*ptr(), reinterpret_cast<const uint8_t *>(code));
  }

  /**
   * @brief Проверить целостность данных.
   *
   * Проверяет соответствие данных, скормленных режиму выработки имитовставки, данной имитовставке.
   *
   * @param code — имитовставка
   *
   * @returns true, в случае соответствия, false в случае несоответствия
   */
  bool verify(const std::array<uint8_t, MAGMA_MAC_RESULT_SIZE> &code) const noexcept
  {
    return magma_mac_verify(*ptr(), code.data());
  }

  /**
   * @brief Сбросить состояние режима выработки имитовставки.
   *
   * Заменяет состояние режима выработки имитовставки новым состоянием.
   */
  void reset() noexcept
  {
    magma_mac_reset(*ptr());
  }

private:
  MagmaMac *ptr() const noexcept
  {
    return const_cast<MagmaMac *>(reinterpret_cast<const MagmaMac *>(&data_));
  }

  MagmaMac *ptr() noexcept
  {
    return reinterpret_cast<MagmaMac *>(&data_);
  }

  aligned_storage_t<MAGMA_MAC_SIZE, MAGMA_MAC_ALIGNMENT> data_;
};

/**
 * Потоковый шифр с длиной блока 128 бит в режиме гаммирования с обратной связью по выходу
 * с размером синхропосылки 16 байт. Определен в ГОСТ Р 34.13-2015 и ГОСТ 34.13-2018.
 *
 * \see #KuznyechikOfb
 */
class KuznyechikOfb_t
{
public:
  KuznyechikOfb_t(const KuznyechikOfb_t &) = default;
  KuznyechikOfb_t &operator=(const KuznyechikOfb_t &) = default;
  KuznyechikOfb_t(KuznyechikOfb_t &&) = default;
  KuznyechikOfb_t &operator=(KuznyechikOfb_t &&) = default;

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по выходу
   * с использованием ключа `key` и синхропосылки `nonce`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   * @param nonce — ненулевой указатель на массив длинной #KUZNYECHIK_OFB_NONCE_SIZE байт
   */
  KuznyechikOfb_t(const uint8_t key[SYMMETRIC_KEY_SIZE], const uint8_t nonce[KUZNYECHIK_OFB_NONCE_SIZE]) noexcept
      : data_{}
  {
    kuznyechik_ofb_new(*ptr(), key, nonce);
  }

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по выходу
   * с использованием ключа `key` и синхропосылки `nonce`.
   *
   * @param key — ключ
   * @param nonce — синхропосылки
   */
  KuznyechikOfb_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key, const std::array<uint8_t, KUZNYECHIK_OFB_NONCE_SIZE> &nonce) noexcept
      : data_{}
  {
    kuznyechik_ofb_new(*ptr(), key.data(), nonce.data());
  }

  ~KuznyechikOfb_t() noexcept
  {
    kuznyechik_ofb_drop(*ptr());
  }

  /**
   * @brief Гаммировать сообщение ключевым потоком шифра.
   *
   * Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
   * шифра.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   */
  void apply_keystream(uint8_t *buf, uintptr_t buf_len) noexcept
  {
    kuznyechik_ofb_apply_keystream(*ptr(), buf, buf_len);
  }

  /**
   * @brief Гаммировать сообщение ключевым потоком шифра.
   *
   * Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
   * шифра.
   *
   * @param buf — сообщение
   */
  void apply_keystream(std::vector<uint8_t> &buf) noexcept
  {
    apply_keystream(buf.data(), buf.size());
  }

private:
  KuznyechikOfb *ptr() noexcept
  {
    return reinterpret_cast<KuznyechikOfb *>(&data_);
  }

  aligned_storage_t<KUZNYECHIK_OFB_SIZE, KUZNYECHIK_OFB_ALIGNMENT> data_;
};

/**
 * Потоковый шифр с длиной блока 64 бит в режиме гаммирования с обратной связью по выходу
 * с размером синхропосылки 8 байт. Определен в ГОСТ Р 34.13-2015 и ГОСТ 34.13-2018.
 *
 * \see #MagmaOfb
 */
class MagmaOfb_t
{
public:
  MagmaOfb_t(const MagmaOfb_t &) = default;
  MagmaOfb_t &operator=(const MagmaOfb_t &) = default;
  MagmaOfb_t(MagmaOfb_t &&) = default;
  MagmaOfb_t &operator=(MagmaOfb_t &&) = default;

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по выходу
   * с использованием ключа `key` и синхропосылки `nonce`.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param key — ненулевой указатель на массив длинной #SYMMETRIC_KEY_SIZE байт
   * @param nonce — ненулевой указатель на массив длинной #MAGMA_OFB_NONCE_SIZE байт
   */
  MagmaOfb_t(const uint8_t key[SYMMETRIC_KEY_SIZE], const uint8_t nonce[MAGMA_OFB_NONCE_SIZE]) noexcept
      : data_{}
  {
    magma_ofb_new(*ptr(), key, nonce);
  }

  /**
   * @brief Создать контекст потокового шифра.
   *
   * Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по выходу
   * с использованием ключа `key` и синхропосылки `nonce`.
   *
   * @param key — ключ
   * @param nonce — синхропосылка
   */
  MagmaOfb_t(const std::array<uint8_t, SYMMETRIC_KEY_SIZE> &key, const std::array<uint8_t, MAGMA_OFB_NONCE_SIZE> &nonce) noexcept
      : data_{}
  {
    magma_ofb_new(*ptr(), key.data(), nonce.data());
  }

  ~MagmaOfb_t() noexcept
  {
    magma_ofb_drop(*ptr());
  }

  /**
   * @brief Гаммировать сообщение ключевым потоком шифра.
   *
   * Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
   * шифра.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param buf — ненулевой указатель на массив байт
   * @param buf_len — длина буфера (в байтах)
   */
  void apply_keystream(uint8_t *buf, uintptr_t buf_len) noexcept
  {
    magma_ofb_apply_keystream(*ptr(), buf, buf_len);
  }

  /**
   * @brief Гаммировать сообщение ключевым потоком шифра.
   *
   * Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
   * шифра.
   *
   * @param buf — сообщение
   */
  void apply_keystream(std::vector<uint8_t> &buf) noexcept
  {
    apply_keystream(buf.data(), buf.size());
  }

private:
  MagmaOfb *ptr() noexcept
  {
    return reinterpret_cast<MagmaOfb *>(&data_);
  }

  aligned_storage_t<MAGMA_OFB_SIZE, MAGMA_OFB_ALIGNMENT> data_;
};

/**
 * Идентификатор эллиптических кривых, для алгоритмов формирования, проверки электронной цифровой
 * подписи и выработки общих ключей. Определены в ГОСТ 34.10-2012 и ГОСТ 34.10-2018.
 */
enum class Gost3410_2012_512 : uintptr_t
{
  /**
   * Идентификатор кривой `id-tc26-gost-3410-12-512-paramSetA`, заданной в Р 1323565.1.024-2019.
   */
  A = GOST3410_2012_512A,

  /**
   * Идентификатор кривой `id-tc26-gost-3410-12-512-paramSetB`, заданной в Р 1323565.1.024-2019.
   */
  B = GOST3410_2012_512B,

  /**
   * Идентификатор кривой `id-tc26-gost-3410-12-512-paramSetC`, заданной в Р 1323565.1.024-2019.
   */
  C = GOST3410_2012_512C,
};

template <Gost3410_2012_512 CurveID>
class Ec512PrivateKey_t;

/**
 * Открытый ключ эллиптических кривых.
 * Специализируется перечислением #Gost3410_2012_512.
 *
 * \see #Ec512
 */
template <Gost3410_2012_512 CurveID>
class Ec512PublicKey_t
{
  friend class Ec512PrivateKey_t<CurveID>;
  const static Ec512 cid = static_cast<Ec512>(CurveID);
  Ec512PublicKey_t<CurveID>() = default;

public:

  Ec512PublicKey_t<CurveID>(const Ec512PublicKey_t<CurveID> &) = default;
  Ec512PublicKey_t<CurveID> &operator=(const Ec512PublicKey_t<CurveID> &) = default;
  Ec512PublicKey_t<CurveID>(Ec512PublicKey_t<CurveID> &&) noexcept = default;
  Ec512PublicKey_t<CurveID> &operator=(Ec512PublicKey_t<CurveID> &&) noexcept = default;
  ~Ec512PublicKey_t<CurveID>() = default;

  /**
   * @brief Десериалировать открытый ключ.
   *
   * Десериализует открытый ключ из представления ГОСТ 34.10-2012 раздел 5.2 во внутреннее представление для заданной эллиптической кривой.
   *
   * @param bytes — сериализованное представление
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_PUBLIC_KEY_BYTES в случае ошибки
   *
   * @returns открытый ключ
   */
  static Ec512PublicKey_t<CurveID> from_bytes(const uint8_t bytes[EC_512_DOUBLE_WIDTH], intptr_t &error_code) noexcept
  {
    Ec512PublicKey_t<CurveID> result;
    error_code = ec512_public_key_from_bytes(cid, bytes, *result.ptr());
    return result;
  }

  /**
   * @brief Десериалировать открытый ключ.
   *
   * Десериализует открытый ключ из представления ГОСТ 34.10-2012 раздел 5.2/ГОСТ 34.10-2018 раздел 5.3 во внутреннее представление для заданной эллиптической кривой.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param bytes — сериализованное представление
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_PUBLIC_KEY_BYTES в случае ошибки
   *
   * @returns открытый ключ
   */
  static Ec512PublicKey_t<CurveID> from_bytes(const std::array<uint8_t, EC_512_DOUBLE_WIDTH> &bytes, intptr_t &error_code) noexcept
  {
    return from_bytes(bytes.data(), error_code);
  }

  /**
   * @brief Сериализовать открытый ключ.
   *
   * Сериализует открытый ключ из внутреннего представления в представление ГОСТ 34.10-2012 раздел 5.2/ГОСТ 34.10-2018 раздел 5.3 для заданной эллиптической кривой.
   *
   * @returns сериализованное представление
   */
  std::array<uint8_t, EC_512_DOUBLE_WIDTH> to_bytes() const noexcept
  {
    std::array<uint8_t, EC_512_DOUBLE_WIDTH> bytes {};
    ec512_public_key_to_bytes(cid, *ptr(), bytes.data());
    return bytes;
  }

  /**
   * @brief Проверить подпись хэша сообщения.
   *
   * Проверяет подпись хэша сообщения с помощью публичной ключа для заданной эллиптической кривой.
   * По ГОСТу хэш сообщения должен быть вычислен с помощью функции Стрибог.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param hash — хэш сообщения
   * @param signature — подпись сообщения
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_SIGNATURE, в случае ошибки
   */
  intptr_t verify_hash(const uint8_t hash[STREEBOG512_RESULT_SIZE],
                       const uint8_t signature[EC_512_DOUBLE_WIDTH]) const noexcept
  {
    return ec512_dsa_verify_hash(cid, *ptr(), hash, signature);
  }

  /**
   * @brief Проверить подпись хэша сообщения.
   *
   * Проверяет подпись хэша сообщения с помощью публичной ключа для заданной эллиптической кривой.
   * По ГОСТу хэш сообщения должен быть вычислен с помощью функции Стрибог.
   *
   * @param hash — хэш сообщения
   * @param signature — подпись сообщения
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_SIGNATURE, в случае ошибки
   */
  intptr_t verify_hash(const std::array<uint8_t, STREEBOG512_RESULT_SIZE> &hash,
                       const std::array<uint8_t, EC_512_DOUBLE_WIDTH> &signature) const noexcept
  {
    return verify_hash(hash.data(), signature.data());
  }

  /**
   * @brief Проверить подпись сообщения.
   *
   * Проверяет подпись сообщения с помощью публичной ключа для заданной эллиптической кривой.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param msg — сообщение
   * @param msg_len — длина сообщения
   * @param signature — подпись сообщения
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_SIGNATURE, в случае ошибки
   */
  intptr_t verify(
      const uint8_t *msg,
      uintptr_t msg_len,
      const uint8_t signature[EC_512_DOUBLE_WIDTH]) const noexcept
  {
    return ec512_dsa_verify(cid, *ptr(), msg, msg_len, signature);
  }

  /**
   * @brief Проверить подпись сообщения.
   *
   * Проверяет подпись сообщения с помощью публичной ключа для заданной эллиптической кривой.
   *
   * @param msg — сообщение
   * @param signature — подпись сообщения
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_SIGNATURE, в случае ошибки
   */
  intptr_t verify(const std::vector<uint8_t> &msg, const std::array<uint8_t, EC_512_DOUBLE_WIDTH> &signature) const noexcept
  {
    return verify(msg.data(), msg.size(), signature.data());
  }

private:
  Ec512PublicKey *ptr() const noexcept
  {
    return const_cast<Ec512PublicKey *>(reinterpret_cast<const Ec512PublicKey *>(&data_));
  }

  aligned_storage_t<EC_512_PUBLIC_KEY_SIZE, EC_512_PUBLIC_KEY_ALIGNMENT> data_;
};

/**
 * Закрытый ключ эллиптических кривых.
 * Специализируется перечислением #Gost3410_2012_512.
 *
 * \see #Ec512
 */
template <Gost3410_2012_512 CurveID>
class Ec512PrivateKey_t
{
  const static Ec512 cid = static_cast<Ec512>(CurveID);
  Ec512PrivateKey_t<CurveID>() = default;

public:

  Ec512PrivateKey_t<CurveID>(const Ec512PrivateKey_t<CurveID> &) = default;
  Ec512PrivateKey_t<CurveID> &operator=(const Ec512PrivateKey_t<CurveID> &) = default;
  Ec512PrivateKey_t<CurveID>(Ec512PrivateKey_t<CurveID> &&) noexcept = default;
  Ec512PrivateKey_t<CurveID> &operator=(Ec512PrivateKey_t<CurveID> &&) noexcept = default;

  ~Ec512PrivateKey_t<CurveID>()
  {
    ec512_private_key_drop(*ptr());
  }

  /**
   * @brief Создать закрытый ключ.
   *
   * Создаёт закрытый ключ для заданной эллиптической кривой.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param error_code код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   */
  Ec512PrivateKey_t<CurveID>(const uint8_t rand_bytes[EC_512_WIDTH], intptr_t &error_code) noexcept
      : data_{}
  {
    error_code = ec512_gen_private_key(cid, *ptr(), rand_bytes);
  }

  /**
   * @brief Создать закрытый ключ.
   *
   * Создаёт закрытый ключ для заданной эллиптической кривой.
   *
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param error_code код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   */
  explicit Ec512PrivateKey_t<CurveID>(const std::array<uint8_t, EC_512_WIDTH> &rand_bytes, intptr_t &error_code) noexcept
      : data_{}
  {
    error_code = ec512_gen_private_key(cid, *ptr(), rand_bytes.data());
  }

  /**
   * @brief Получить открытый ключ.
   *
   * Получает открытый ключ из закрытого ключа для заданной эллиптической кривой.
   *
   * @returns открытый ключ
   */
  Ec512PublicKey_t<CurveID> get_public_key() const noexcept
  {
    Ec512PublicKey_t<CurveID> public_key;
    ec512_get_public_key(cid, *ptr(), *public_key.ptr());
    return public_key;
  }

  /**
   * @brief Десериалировать закрытый ключ.
   *
   * Десериализует закрытый ключ из представления ГОСТ 34.10-2012 раздел 5.2/ГОСТ 34.10-2018 раздел 5.3 во внутреннее представление для заданной эллиптической кривой.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param bytes — сериализованное представление
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   *
   * @returns десериализованный закрытый ключ
   */
  static Ec512PrivateKey_t<CurveID> from_bytes(const uint8_t bytes[EC_512_WIDTH], intptr_t &error_code) noexcept
  {
    Ec512PrivateKey_t<CurveID> result;
    error_code = ec512_private_key_from_bytes(cid, bytes, *result.ptr());
    return result;
  }

  /**
   * @brief Десериалировать закрытый ключ.
   *
   * Десериализует закрытый ключ из представления ГОСТ 34.10-2012 раздел 5.2/ГОСТ 34.10-2018 раздел 5.3 во внутреннее представление для заданной эллиптической кривой.
   *
   * @param rand_bytes — сериализованное представление
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   *
   * @returns десериализованный закрытый ключ
   */
  static Ec512PrivateKey_t<CurveID> from_bytes(const std::array<uint8_t, EC_512_WIDTH> &rand_bytes, intptr_t &error_code) noexcept
  {
    return from_bytes(rand_bytes.data(), error_code);
  }

  /**
   * @brief Сериализовать закрытый ключ.
   *
   * Сериализует закрытый ключ из внутреннего представления в представление ГОСТ 34.10-2012 раздел 5.2/ГОСТ 34.10-2018 раздел 5.3 для заданной эллиптической кривой.
   *
   * @returns сериализованный закрытый
   */
  std::array<uint8_t, EC_512_WIDTH> to_bytes() const noexcept
  {
    std::array<uint8_t, EC_512_WIDTH> bytes {};
    ec512_private_key_to_bytes(cid, *ptr(), bytes.data());
    return bytes;
  }

  /**
   * @brief Подписать хэш сообщения.
   *
   * Подписывает хэш сообщения с помощью закрытого ключа для заданной эллиптической кривой.
   * По ГОСТу хэш сообщения должен быть вычислен с помощью функции Стрибог.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param hash — хэш сообщения
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param signature — результирующая подпись
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES, в случае ошибки
   */
  intptr_t sign_hash(const uint8_t hash[STREEBOG512_RESULT_SIZE],
                     const uint8_t rand_bytes[EC_512_WIDTH],
                     uint8_t signature[EC_512_DOUBLE_WIDTH]) const noexcept
  {
    return ec512_dsa_sign_hash(cid, *ptr(), hash, rand_bytes, signature);
  }

  /**
   * @brief Подписать хэш сообщения.
   *
   * Подписывает хэш сообщения с помощью закрытого ключа для заданной эллиптической кривой.
   * По ГОСТу хэш сообщения должен быть вычислен с помощью функции Стрибог.
   *
   * @param hash — хэш сообщения
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   *
   * @returns результирующая подпись
   */
  std::array<uint8_t, EC_512_DOUBLE_WIDTH> sign_hash(
      const std::array<uint8_t, STREEBOG512_RESULT_SIZE> &hash,
      const std::array<uint8_t, EC_512_WIDTH> &rand_bytes,
      intptr_t &error_code) const noexcept
  {
    std::array<uint8_t, EC_512_DOUBLE_WIDTH> signature {};
    error_code = sign_hash(hash.data(), rand_bytes.data(), signature.data());
    return signature;
  }

  /**
   * @brief Подписать сообщение.
   *
   * Подписывает сообщение с помощью закрытого ключа для заданной эллиптической кривой.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param msg — указатель на сообщение
   * @param msg_len — длина сообщения
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param signature — результирующая подпись
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES, в случае ошибки
   */
  intptr_t sign(
      const uint8_t *msg,
      uintptr_t msg_len,
      const uint8_t rand_bytes[EC_512_WIDTH],
      uint8_t signature[EC_512_DOUBLE_WIDTH]) const noexcept
  {
    return ec512_dsa_sign(cid, *ptr(), msg, msg_len, rand_bytes, signature);
  }

  /**
   * @brief Подписать сообщение.
   *
   * Подписывает сообщение с помощью закрытого ключа для заданной эллиптической кривой.
   *
   * @param msg — указатель на сообщение
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param signature — результирующая подпись
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES, в случае ошибки
   */
  intptr_t sign(const std::vector<uint8_t> &msg, const std::array<uint8_t, EC_512_WIDTH> &rand_bytes, std::array<uint8_t, EC_512_DOUBLE_WIDTH> &signature) const noexcept
  {
    return sign(msg.data(), msg.size(), rand_bytes.data(), signature.data());
  }

  /**
   * @brief Подписать сообщение.
   *
   * Подписывает сообщение с помощью закрытого ключа для заданной эллиптической кривой.
   *
   * @param msg — указатель на сообщение
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   *
   * @returns сигнатуру
   */
  std::array<uint8_t, EC_512_DOUBLE_WIDTH> sign(const std::vector<uint8_t> &msg, const std::array<uint8_t, EC_512_WIDTH> &rand_bytes, intptr_t &error_code) const noexcept
  {
    std::array<uint8_t, EC_512_DOUBLE_WIDTH> signature {};
    error_code = sign(msg, rand_bytes, signature);
    return signature;
  }

  /**
   * @brief Выработать общий ключ.
   *
   * Вырабатывает алгоритм выработки ключей обмена длиной 256 бит, используя 512-битные
   * эллиптические кривые.
   *
   * @param other_pk — чужой открытый ключ
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_VKO_UKM в случае ошибки
   *
   * @returns общий ключ
   */
  std::array<uint8_t, EC_256_WIDTH> vko256(
      const Ec512PublicKey_t<CurveID> &other_pk,
      intptr_t &error_code) const noexcept
  {
    std::array<uint8_t, EC_256_WIDTH> out {};
    error_code = ec512_vko256(cid, *ptr(), *other_pk.ptr(), nullptr, out.data());
    return out;
  }

  /**
   * @brief Выработать общий ключ.
   *
   * Вырабатывает алгоритм выработки ключей обмена длиной 256 бит, используя 512-битные
   * эллиптические кривые.
   *
   * Если в `ukm` передан нулевой указатель, то используется значение по
   * умолчанию равное единице. Если указатель не нулевой, то значение на
   * которое он указывает не должно быть равным нулю по модулю простого
   * числа эллиптической кривой. В случае передачи такого значения
   * возвращается код ошибки #EC_ERR_VKO_UKM. Если используемый приватный
   * или публичный ключ статичны (т.е. используются неэфемерные ключи), то
   * рекомендуется использовать `ukm`, содержащий не менее 128 переменных бит.
   *
   * \warning Передача некорректного указателя приводит к неопределённому поведению.
   *
   * @param other_pk — чужой открытый ключ
   * @param ukm — указатель на пользовательский ключевой материал (опционально)
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_VKO_UKM в случае ошибки
   *
   * @returns общий ключ
   */
  std::array<uint8_t, EC_256_WIDTH> vko256(
      const Ec512PublicKey_t<CurveID> &other_pk,
      intptr_t &error_code,
      const uint8_t ukm[EC_512_WIDTH]) const noexcept
  {
    std::array<uint8_t, EC_256_WIDTH> out {};
    error_code = ec512_vko256(cid, *ptr(), *other_pk.ptr(), ukm, out.data());
    return out;
  }

  /**
   * @brief Выработать общий ключ.
   *
   * Вырабатывает алгоритм выработки ключей обмена длиной 512 бит, используя 512-битные
   * эллиптические кривые.
   *
   * @param other_pk — чужой открытый ключ
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_VKO_UKM в случае ошибки
   *
   * @returns общий ключ
   */
  std::array<uint8_t, EC_512_WIDTH> vko512(
      const Ec512PublicKey_t<CurveID> &other_pk,
      intptr_t &error_code) const noexcept
  {
    std::array<uint8_t, EC_512_WIDTH> out {};
    error_code = ec512_vko512(cid, *ptr(), *other_pk.ptr(), nullptr, out.data());
    return out;
  }

  /**
   * @brief Выработать общий ключ.
   *
   * Вырабатывает алгоритм выработки ключей обмена длиной 512 бит, используя 512-битные
   * эллиптические кривые.
   *
   * Если в `ukm` передан нулевой указатель, то используется значение по
   * умолчанию равное единице. Если указатель не нулевой, то значение на
   * которое он указывает не должно быть равным нулю по модулю простого
   * числа эллиптической кривой. В случае передачи такого значения
   * возвращается код ошибки #EC_ERR_VKO_UKM. Если используемый приватный
   * или публичный ключ статичны (т.е. используются неэфемерные ключи), то
   * рекомендуется использовать `ukm`, содержащий не менее 128 переменных бит.
   *
   * \warning Передача некорректного указателя приводит к неопределённому поведению.
   *
   * @param other_pk — чужой открытый ключ
   * @param ukm — указатель на пользовательский ключевой материал (опционально)
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_VKO_UKM в случае ошибки
   *
   * @returns общий ключ
   */
  std::array<uint8_t, EC_512_WIDTH> vko512(
      const Ec512PublicKey_t<CurveID> &other_pk,
      intptr_t &error_code,
      const uint8_t ukm[EC_512_WIDTH]) const noexcept
  {
    std::array<uint8_t, EC_512_WIDTH> out {};
    error_code = ec512_vko512(cid, *ptr(), *other_pk.ptr(), ukm, out.data());
    return out;
  }

private:
  Ec512PrivateKey *ptr() const noexcept
  {
    return const_cast<Ec512PrivateKey *>(reinterpret_cast<const Ec512PrivateKey *>(&data_));
  }

  aligned_storage_t<EC_512_WIDTH, EC_512_PRIVATE_KEY_ALIGNMENT> data_ {};
};

/**
 * Идентификатор эллиптических кривых, для алгоритмов формирования, проверки электронной цифровой
 * подписи и выработки общих ключей. Определены в ГОСТ 34.10-2012 и ГОСТ 34.10-2018.
 */
enum class Gost3410_2012_256 : uintptr_t
{
  /**
   * Идентификатор кривой `id-tc26-gost-3410-2012-256-paramSetA`, заданной в Р 1323565.1.024-2019.
   */
  A = GOST3410_2012_256A,

  /**
   * Идентификатор кривой `id-tc26-gost-3410-2012-256-paramSetB`, заданной в Р 1323565.1.024-2019.
   */
  B = GOST3410_2012_256B,

  /**
   * Идентификатор кривой `id-tc26-gost-3410-2012-256-paramSetC`, заданной в Р 1323565.1.024-2019.
   */
  C = GOST3410_2012_256C,

  /**
   * Идентификатор кривой `id-tc26-gost-3410-2012-256-paramSetD`, заданной в Р 1323565.1.024-2019.
   */
  D = GOST3410_2012_256D
};

template <Gost3410_2012_256 CurveID>
class Ec256PrivateKey_t;

/**
 * Открытый ключ эллиптических кривых.
 * Специализируется перечислением #Gost3410_2012_256.
 *
 * \see #Ec256
 */
template <Gost3410_2012_256 CurveID>
class Ec256PublicKey_t
{
  friend class Ec256PrivateKey_t<CurveID>;
  const static Ec256 cid = static_cast<Ec256>(CurveID);
  Ec256PublicKey_t<CurveID>() = default;

public:

  Ec256PublicKey_t<CurveID>(const Ec256PublicKey_t<CurveID> &) = default;
  Ec256PublicKey_t<CurveID> &operator=(const Ec256PublicKey_t<CurveID> &) = default;
  Ec256PublicKey_t<CurveID>(Ec256PublicKey_t<CurveID> &&) noexcept = default;
  Ec256PublicKey_t<CurveID> &operator=(Ec256PublicKey_t<CurveID> &&) noexcept = default;
  ~Ec256PublicKey_t<CurveID>() = default;

  /**
   * @brief Десериалировать открытый ключ.
   *
   * Десериализует открытый ключ из представления ГОСТ 34.10-2012 раздел 5.2/ГОСТ 34.10-2018 раздел 5.3 во внутреннее представление для заданной эллиптической кривой.
   *
   * @param bytes — сериализованное представление
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_PUBLIC_KEY_BYTES в случае ошибки
   *
   * @returns открытый ключ
   */
  static Ec256PublicKey_t<CurveID> from_bytes(const uint8_t bytes[EC_256_DOUBLE_WIDTH], intptr_t &error_code) noexcept
  {
    Ec256PublicKey_t<CurveID> result;
    error_code = ec256_public_key_from_bytes(cid, bytes, *result.ptr());
    return result;
  }

  /**
   * @brief Десериалировать открытый ключ.
   *
   * Десериализует открытый ключ из представления ГОСТ 34.10-2012 раздел 5.2/ГОСТ 34.10-2018 раздел 5.3 во внутреннее представление для заданной эллиптической кривой.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param bytes — сериализованное представление
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_PUBLIC_KEY_BYTES в случае ошибки
   *
   * @returns открытый ключ
   */
  static Ec256PublicKey_t<CurveID> from_bytes(const std::array<uint8_t, EC_256_DOUBLE_WIDTH> &bytes, intptr_t &error_code) noexcept
  {
    return from_bytes(bytes.data(), error_code);
  }

  /**
   * @brief Сериализовать открытый ключ.
   *
   * Сериализует открытый ключ из внутреннего представления в представление ГОСТ 34.10-2012 раздел 5.2/ГОСТ 34.10-2018 раздел 5.3 для заданной эллиптической кривой.
   *
   * @returns сериализованное представление
   */
  std::array<uint8_t, EC_256_DOUBLE_WIDTH> to_bytes() const noexcept
  {
    std::array<uint8_t, EC_256_DOUBLE_WIDTH> bytes {};
    ec256_public_key_to_bytes(cid, *ptr(), bytes.data());
    return bytes;
  }

  /**
   * @brief Проверить подпись хэша сообщения.
   *
   * Проверяет подпись хэша сообщения с помощью публичной ключа для заданной эллиптической кривой.
   * По ГОСТу хэш сообщения должен быть вычислен с помощью функции Стрибог.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param hash — хэш сообщения
   * @param signature — подпись сообщения
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_SIGNATURE, в случае ошибки
   */
  intptr_t verify_hash(const uint8_t hash[STREEBOG256_RESULT_SIZE],
                       const uint8_t signature[EC_256_DOUBLE_WIDTH]) const noexcept
  {
    return ec256_dsa_verify_hash(cid, *ptr(), hash, signature);
  }

  /**
   * @brief Проверить подпись хэша сообщения.
   *
   * Проверяет подпись хэша сообщения с помощью публичной ключа для заданной эллиптической кривой.
   * По ГОСТу хэш сообщения должен быть вычислен с помощью функции Стрибог.
   *
   * @param hash — хэш сообщения
   * @param signature — подпись сообщения
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_SIGNATURE, в случае ошибки
   */
  intptr_t verify_hash(const std::array<uint8_t, STREEBOG256_RESULT_SIZE> &hash,
                       const std::array<uint8_t, EC_256_DOUBLE_WIDTH> &signature) const noexcept
  {
    return verify_hash(hash.data(), signature.data());
  }

  /**
   * @brief Проверить подпись сообщения.
   *
   * Проверяет подпись сообщения с помощью публичной ключа для заданной эллиптической кривой.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param msg — сообщение
   * @param msg_len — длина сообщения
   * @param signature — подпись сообщения
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_SIGNATURE, в случае ошибки
   */
  intptr_t verify(
      const uint8_t *msg,
      uintptr_t msg_len,
      const uint8_t signature[EC_256_DOUBLE_WIDTH]) const noexcept
  {
    return ec256_dsa_verify(cid, *ptr(), msg, msg_len, signature);
  }

  /**
   * @brief Проверить подпись сообщения.
   *
   * Проверяет подпись сообщения с помощью публичной ключа для заданной эллиптической кривой.
   *
   * @param msg — сообщение
   * @param signature — подпись сообщения
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_SIGNATURE, в случае ошибки
   */
  intptr_t verify(const std::vector<uint8_t> &msg, const std::array<uint8_t, EC_256_DOUBLE_WIDTH> &signature) const noexcept
  {
    return verify(msg.data(), msg.size(), signature.data());
  }

private:
  Ec256PublicKey *ptr() const noexcept
  {
    return const_cast<Ec256PublicKey *>(reinterpret_cast<const Ec256PublicKey *>(&data_));
  }

  aligned_storage_t<EC_256_PUBLIC_KEY_SIZE, EC_256_PUBLIC_KEY_ALIGNMENT> data_;
};

/**
 * Закрытый ключ эллиптических кривых.
 * Специализируется перечислением #Gost3410_2012_256.
 *
 * \see #Ec256
 */
template <Gost3410_2012_256 CurveID>
class Ec256PrivateKey_t
{
  const static Ec256 cid = static_cast<Ec256>(CurveID);
  Ec256PrivateKey_t<CurveID>() = default;

public:

  Ec256PrivateKey_t<CurveID>(const Ec256PrivateKey_t<CurveID> &) = default;
  Ec256PrivateKey_t<CurveID> &operator=(const Ec256PrivateKey_t<CurveID> &) = default;
  Ec256PrivateKey_t<CurveID>(Ec256PrivateKey_t<CurveID> &&) noexcept = default;
  Ec256PrivateKey_t<CurveID> &operator=(Ec256PrivateKey_t<CurveID> &&) noexcept = default;

  ~Ec256PrivateKey_t<CurveID>()
  {
    ec256_private_key_drop(*ptr());
  }

  /**
   * @brief Создать закрытый ключ.
   *
   * Создаёт закрытый ключ для заданной эллиптической кривой.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param error_code код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   */
  Ec256PrivateKey_t<CurveID>(const uint8_t rand_bytes[EC_256_WIDTH], intptr_t &error_code) noexcept
      : data_{}
  {
    error_code = ec256_gen_private_key(cid, *ptr(), rand_bytes);
  }

  /**
   * @brief Создать закрытый ключ.
   *
   * Создаёт закрытый ключ для заданной эллиптической кривой.
   *
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param error_code код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   */
  explicit Ec256PrivateKey_t<CurveID>(const std::array<uint8_t, EC_256_WIDTH> &rand_bytes, intptr_t &error_code) noexcept
      : data_{}
  {
    error_code = ec256_gen_private_key(cid, *ptr(), rand_bytes.data());
  }

  /**
   * @brief Получить открытый ключ.
   *
   * Получает открытый ключ из закрытого ключа для заданной эллиптической кривой.
   *
   * @returns открытый ключ
   */
  Ec256PublicKey_t<CurveID> get_public_key() const noexcept
  {
    Ec256PublicKey_t<CurveID> public_key;
    ec256_get_public_key(cid, *ptr(), *public_key.ptr());
    return public_key;
  }

  /**
   * @brief Десериалировать закрытый ключ.
   *
   * Десериализует закрытый ключ из представления ГОСТ 34.10-2012 раздел 5.2/ГОСТ 34.10-2018 раздел 5.3 во внутреннее представление для заданной эллиптической кривой.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param bytes — сериализованное представление
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   *
   * @returns десериализованный закрытый ключ
   */
  static Ec256PrivateKey_t<CurveID> from_bytes(const uint8_t bytes[EC_256_WIDTH], intptr_t &error_code) noexcept
  {
    Ec256PrivateKey_t<CurveID> result;
    error_code = ec256_private_key_from_bytes(cid, bytes, *result.ptr());
    return result;
  }

  /**
   * @brief Десериалировать закрытый ключ.
   *
   * Десериализует закрытый ключ из представления ГОСТ 34.10-2012 раздел 5.2/ГОСТ 34.10-2018 раздел 5.3 во внутреннее представление для заданной эллиптической кривой.
   *
   * @param rand_bytes — сериализованное представление
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   *
   * @returns десериализованный закрытый ключ
   */
  static Ec256PrivateKey_t<CurveID> from_bytes(const std::array<uint8_t, EC_256_WIDTH> &rand_bytes, intptr_t &error_code) noexcept
  {
    return from_bytes(rand_bytes.data(), error_code);
  }

  /**
   * @brief Сериализовать закрытый ключ.
   *
   * Сериализует закрытый ключ из внутреннего представления в представление ГОСТ 34.10-2012 раздел 5.2 для заданной эллиптической кривой.
   *
   * @returns сериализованный закрытый
   */
  std::array<uint8_t, EC_256_WIDTH> to_bytes() const noexcept
  {
    std::array<uint8_t, EC_256_WIDTH> bytes {};
    ec256_private_key_to_bytes(cid, *ptr(), bytes.data());
    return bytes;
  }

  /**
   * @brief Подписать хэш сообщения.
   *
   * Подписывает хэш сообщения с помощью закрытого ключа для заданной эллиптической кривой.
   * По ГОСТу хэш сообщения должен быть вычислен с помощью функции Стрибог.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param hash — хэш сообщения
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param signature — результирующая подпись
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES, в случае ошибки
   */
  intptr_t sign_hash(const uint8_t hash[STREEBOG256_RESULT_SIZE],
                     const uint8_t rand_bytes[EC_256_WIDTH],
                     uint8_t signature[EC_256_DOUBLE_WIDTH]) const noexcept
  {
    return ec256_dsa_sign_hash(cid, *ptr(), hash, rand_bytes, signature);
  }

  /**
   * @brief Подписать хэш сообщения.
   *
   * Подписывает хэш сообщения с помощью закрытого ключа для заданной эллиптической кривой.
   * По ГОСТу хэш сообщения должен быть вычислен с помощью функции Стрибог.
   *
   * @param hash — хэш сообщения
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   *
   * @returns результирующая подпись
   */
  std::array<uint8_t, EC_256_DOUBLE_WIDTH> sign_hash(
      const std::array<uint8_t, STREEBOG256_RESULT_SIZE> &hash,
      const std::array<uint8_t, EC_256_WIDTH> &rand_bytes,
      intptr_t &error_code) const noexcept
  {
    std::array<uint8_t, EC_256_DOUBLE_WIDTH> signature {};
    error_code = sign_hash(hash.data(), rand_bytes.data(), signature.data());
    return signature;
  }

  /**
   * @brief Подписать сообщение.
   *
   * Подписывает сообщение с помощью закрытого ключа для заданной эллиптической кривой.
   *
   * \warning Передача нулевого или некорректного указателя приводит к неопределённому поведению.
   *
   * @param msg — указатель на сообщение
   * @param msg_len — длина сообщения
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param signature — результирующая подпись
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES, в случае ошибки
   */
  intptr_t sign(
      const uint8_t *msg,
      uintptr_t msg_len,
      const uint8_t rand_bytes[EC_256_WIDTH],
      uint8_t signature[EC_256_DOUBLE_WIDTH]) const noexcept
  {
    return ec256_dsa_sign(cid, *ptr(), msg, msg_len, rand_bytes, signature);
  }

  /**
   * @brief Подписать сообщение.
   *
   * Подписывает сообщение с помощью закрытого ключа для заданной эллиптической кривой.
   *
   * @param msg — указатель на сообщение
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param signature — результирующая подпись
   *
   * @returns 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES, в случае ошибки
   */
  intptr_t sign(const std::vector<uint8_t> &msg, const std::array<uint8_t, EC_256_WIDTH> &rand_bytes, std::array<uint8_t, EC_256_DOUBLE_WIDTH> &signature) const noexcept
  {
    return sign(msg.data(), msg.size(), rand_bytes.data(), signature.data());
  }

  /**
   * @brief Подписать сообщение.
   *
   * Подписывает сообщение с помощью закрытого ключа для заданной эллиптической кривой.
   *
   * @param msg — указатель на сообщение
   * @param rand_bytes — набор чисел, сгенерированный случайным образом
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_BAD_RAND_BYTES в случае ошибки
   *
   * @returns сигнатуру
   */
  std::array<uint8_t, EC_256_DOUBLE_WIDTH> sign(const std::vector<uint8_t> &msg, const std::array<uint8_t, EC_256_WIDTH> &rand_bytes, intptr_t &error_code) const noexcept
  {
    std::array<uint8_t, EC_256_DOUBLE_WIDTH> signature {};
    error_code = sign(msg, rand_bytes, signature);
    return signature;
  }

  /**
   * @brief Выработать общий ключ.
   *
   * Вырабатывает 256-битный алгоритм выработки ключей обмена длиной 256 бит, используя 256-битные
   * эллиптические кривые.
   *
   * @param other_pk — чужой открытый ключ
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_VKO_UKM в случае ошибки
   *
   * @returns общий ключ
   */
  std::array<uint8_t, EC_256_WIDTH> vko256(
      const Ec256PublicKey_t<CurveID> &other_pk,
      intptr_t &error_code) const noexcept
  {
    std::array<uint8_t, EC_256_WIDTH> out {};
    error_code = ec256_vko256(cid, *ptr(), *other_pk.ptr(), nullptr, out.data());
    return out;
  }

  /**
   * @brief Выработать общий ключ.
   *
   * Вырабатывает 256-битный алгоритм выработки ключей обмена длиной 256 бит, используя 256-битные
   * эллиптические кривые.
   *
   * Если в `ukm` передан нулевой указатель, то используется значение по
   * умолчанию равное единице. Если указатель не нулевой, то значение на
   * которое он указывает не должно быть равным нулю по модулю простого
   * числа эллиптической кривой. В случае передачи такого значения
   * возвращается код ошибки #EC_ERR_VKO_UKM. Если используемый приватный
   * или публичный ключ статичны (т.е. используются неэфемерные ключи), то
   * рекомендуется использовать `ukm`, содержащий не менее 128 переменных бит.
   *
   * \warning Передача некорректного указателя приводит к неопределённому поведению.
   *
   * @param other_pk — чужой открытый ключ
   * @param ukm — указатель на пользовательский ключевой материал (опционально)
   * @param error_code — код ошибки: 0, в случае успеха, #EC_ERR_CURVE_ID, #EC_ERR_VKO_UKM в случае ошибки
   *
   * @returns общий ключ
   */
  std::array<uint8_t, EC_256_WIDTH> vko256(
      const Ec256PublicKey_t<CurveID> &other_pk,
      intptr_t &error_code,
      const uint8_t ukm[EC_256_WIDTH]) const noexcept
  {
    std::array<uint8_t, EC_256_WIDTH> out {};
    error_code = ec256_vko256(cid, *ptr(), *other_pk.ptr(), ukm, out.data());
    return out;
  }

private:
  Ec256PrivateKey *ptr() const noexcept
  {
    return const_cast<Ec256PrivateKey *>(reinterpret_cast<const Ec256PrivateKey *>(&data_));
  }

  aligned_storage_t<EC_256_WIDTH, EC_256_PRIVATE_KEY_ALIGNMENT> data_ {};
};

#endif /* GOSTCRYPTO_HPP */
