#![no_std]

use rand::{CryptoRng, Rng};
use static_assertions::const_assert_eq;

use core::cmp::Ordering;
use core::hash::{Hash, Hasher};
use core::mem::{zeroed, MaybeUninit};

#[allow(dead_code)]
mod sys;

/// Размер ключа симметричного шифра (в байтах)
pub const SYMMETRIC_KEY_SIZE: usize = sys::SYMMETRIC_KEY_SIZE as _;

/// Размер блока шифра [Magma] (в байтах)
pub const MAGMA_BLOCK_SIZE: usize = sys::MAGMA_BLOCK_SIZE as _;
/// Размер блока шифра [Kuznyechik] (в байтах)
pub const KUZNYECHIK_BLOCK_SIZE: usize = sys::KUZNYECHIK_BLOCK_SIZE as _;

/// Размер синхропосылки [KuznyechikMgm] (в байтах)
pub const KUZNYECHIK_MGM_NONCE_SIZE: usize = sys::KUZNYECHIK_MGM_NONCE_SIZE as _;
/// Размер имитовставки [KuznyechikMgm] (в байтах)
pub const KUZNYECHIK_MGM_TAG_SIZE: usize = sys::KUZNYECHIK_MGM_TAG_SIZE as _;
/// Размер синхропосылки [MagmaCbc] (в байтах)
pub const MAGMA_CBC_NONCE_SIZE: usize = sys::MAGMA_CBC_NONCE_SIZE as _;
/// Размер синхропосылки [KuznyechikCbc] (в байтах)
pub const KUZNYECHIK_CBC_NONCE_SIZE: usize = sys::KUZNYECHIK_CBC_NONCE_SIZE as _;
/// Размер синхропосылки [MagmaCfb] (в байтах)
pub const MAGMA_CFB_NONCE_SIZE: usize = sys::MAGMA_CFB_NONCE_SIZE as _;
/// Размер синхропосылки [KuznyechikCfb] (в байтах)
pub const KUZNYECHIK_CFB_NONCE_SIZE: usize = sys::KUZNYECHIK_CFB_NONCE_SIZE as _;
/// Размер синхропосылки [MagmaCtr] (в байтах)
pub const MAGMA_CTR_NONCE_SIZE: usize = sys::MAGMA_CTR_NONCE_SIZE as _;
/// Размер синхропосылки [KuznyechikCtr] (в байтах)
pub const KUZNYECHIK_CTR_NONCE_SIZE: usize = sys::KUZNYECHIK_CTR_NONCE_SIZE as _;
/// Размер синхропосылки [MagmaOfb] (в байтах)
pub const MAGMA_OFB_NONCE_SIZE: usize = sys::MAGMA_OFB_NONCE_SIZE as _;
/// Размер синхропосылки [KuznyechikOfb] (в байтах)
pub const KUZNYECHIK_OFB_NONCE_SIZE: usize = sys::KUZNYECHIK_OFB_NONCE_SIZE as _;

/// Длина закрытого ключа и хэша сообщения кривых Ec256 (в байтах)
pub const EC_256_WIDTH: usize = sys::EC_256_WIDTH as _;
/// Длина сериализованного открытого ключа и подписи кривых Ec256 (в байтах)
pub const EC_256_DOUBLE_WIDTH: usize = sys::EC_256_DOUBLE_WIDTH as _;
/// Длина несериализованного открытого ключа кривых Ec256 (в байтах)
pub const EC_256_PUBLIC_KEY_SIZE: usize = sys::EC_256_PUBLIC_KEY_SIZE as _;
/// Длина закрытого ключа и хэша сообщения кривых Ec512 (в байтах)
pub const EC_512_WIDTH: usize = sys::EC_512_WIDTH as _;
/// Длина сериализованного открытого ключа и подписи кривых Ec512 (в байтах)
pub const EC_512_DOUBLE_WIDTH: usize = sys::EC_512_DOUBLE_WIDTH as _;
/// Длина несериализованного открытого ключа кривых Ec512 (в байтах)
pub const EC_512_PUBLIC_KEY_SIZE: usize = sys::EC_512_PUBLIC_KEY_SIZE as _;

/// Длина хэш-кода [Streebog256] (в байтах)
pub const STREEBOG256_RESULT_SIZE: usize = sys::STREEBOG256_RESULT_SIZE as _;
/// Длина хэш-кода [Streebog512] (в байтах)
pub const STREEBOG512_RESULT_SIZE: usize = sys::STREEBOG512_RESULT_SIZE as _;
/// Длина имитовставки [MagmaMac] (в байтах)
pub const MAGMA_MAC_RESULT_SIZE: usize = sys::MAGMA_MAC_RESULT_SIZE as _;
/// Длина имитовставки [KuznyechikMac] (в байтах)
pub const KUZNYECHIK_MAC_RESULT_SIZE: usize = sys::KUZNYECHIK_MAC_RESULT_SIZE as _;

mod private {
    pub trait Sealed {}
}

/// Кривая Ec256
pub trait Ec256Curve: private::Sealed {
    /// Идентификатор кривой
    const CURVE_ID: usize;
}

/// Кривая Ec512
pub trait Ec512Curve: private::Sealed {
    /// Идентификатор кривой
    const CURVE_ID: usize;
}

macro_rules! define_struct {
    ($(#[$meta:meta])* $name:ident, $size:path, $align:tt, $align_const:path) => {
        const_assert_eq!($align, $align_const);

        $(#[$meta])*
        #[repr(C, align($align))]
        pub struct $name {
            inner: sys::$name
        }

        impl $name {
            fn as_mut_ptr(&mut self) -> *mut u8 {
                self.inner.as_mut_ptr()
            }
        }

        impl Clone for $name {
            fn clone(&self) -> Self {
                Self { inner: self.inner.clone() }
            }
        }
    }
}

macro_rules! ec_struct {
    ($(#[$meta:meta])* $name:ident, $align:tt, $align_const:path, $curve_id:ident) => {
        const_assert_eq!($align, $align_const);

        $(#[$meta])*
        #[repr(C, align($align))]
        pub struct $name<C: $curve_id> {
            inner: sys::$name,
            curve_id: ::core::marker::PhantomData<C>,
        }

        impl<C: $curve_id> $name<C> {
            fn zeroed() -> Self {
                Self {
                    inner: [0; ::core::mem::size_of::<sys::$name>()],
                    curve_id: Default::default(),
                }
            }

            fn as_ptr(&self) -> *const u8 {
                self.inner.as_ptr()
            }

            fn as_mut_ptr(&mut self) -> *mut u8 {
                self.inner.as_mut_ptr()
            }
        }

        impl<C: $curve_id> Clone for $name<C> {
            fn clone(&self) -> Self {
                Self {
                    inner: self.inner.clone(),
                    curve_id: self.curve_id,
                }
            }
        }

        impl<C: $curve_id> PartialEq for $name<C> {
            fn eq(&self, other: &Self) -> bool {
                self.inner == other.inner
            }
        }

        impl<C: $curve_id> Eq for $name<C> {}

        impl<C: $curve_id> PartialOrd for $name<C> {
            fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
                Some(self.inner.cmp(&other.inner))
            }
        }

        impl<C: $curve_id> Ord for $name<C> {
            fn cmp(&self, other: &Self) -> Ordering {
                self.inner.cmp(&other.inner)
            }
        }

        impl<C: $curve_id> Hash for $name<C> {
            fn hash<H: Hasher>(&self, state: &mut H) {
                self.inner.hash(state);
            }
        }
    }
}

macro_rules! impl_drop {
    ($ty:ident, $drop:path) => {
        impl Drop for $ty {
            fn drop(&mut self) {
                unsafe { $drop(self.as_mut_ptr()) }
            }
        }
    };
}

define_struct!(
    /// Блочный шифр с длиной блока 128 бит в режиме работы блочных шифров, реализующим
    /// аутентифицирующее шифрование с присоединёнными данными (AEAD). Определен в Р 1323565.1.026–2019.
    KuznyechikMgm,
    sys::KUZNYECHIK_MGM_SIZE,
    1,
    sys::KUZNYECHIK_MGM_ALIGNMENT
);
define_struct!(
    /// Блочный шифр с длиной блока 64 бит в режиме простой замены с зацеплением
    /// с размером синхропосылки 8 байт. Определен в ГОСТ 34.13-2015.
    MagmaCbc,
    sys::MAGMA_CBC_SIZE,
    4,
    sys::MAGMA_CBC_ALIGNMENT
);
define_struct!(
    /// Блочный шифр с длиной блока 128 бит в режиме простой замены с зацеплением
    /// с размером синхропосылки 16 байт. Определен в ГОСТ 34.13-2015.
    KuznyechikCbc,
    sys::KUZNYECHIK_CBC_SIZE,
    1,
    sys::KUZNYECHIK_CBC_ALIGNMENT
);
define_struct!(
    /// Потоковый шифр с длиной блока 64 бит в режиме гаммирования с обратной связью по шифртексту
    /// с размером синхропосылки 8 байт. Определен в ГОСТ 34.13-2015.
    MagmaCfb,
    sys::MAGMA_CFB_SIZE,
    4,
    sys::MAGMA_CFB_ALIGNMENT
);
define_struct!(
    /// Потоковый шифр с длиной блока 128 бит в режиме гаммирования с обратной связью по шифртексту
    /// с размером синхропосылки 16 байт. Определен в ГОСТ 34.13-2015.
    KuznyechikCfb,
    sys::KUZNYECHIK_CFB_SIZE,
    1,
    sys::KUZNYECHIK_CFB_ALIGNMENT
);
define_struct!(
    /// Блочный шифр с длиной блока 64 бит. Определён в ГОСТ Р 34.12-2015 (Магма).
    Magma,
    sys::MAGMA_SIZE,
    4,
    sys::MAGMA_ALIGNMENT
);
define_struct!(
    /// Блочный шифр с длиной блока 128 бит. Определён в ГОСТ Р 34.12-2015 (Кузнечик).
    Kuznyechik,
    sys::KUZNYECHIK_SIZE,
    1,
    sys::KUZNYECHIK_ALIGNMENT
);
define_struct!(
    /// Потоковый шифр с длиной блока 64 бит в режиме гаммирования. Определен в ГОСТ 34.13-2015.
    MagmaCtr,
    sys::MAGMA_CTR_SIZE,
    4,
    sys::MAGMA_CTR_ALIGNMENT
);
define_struct!(
    /// Потоковый шифр с длиной блока 128 бит в режиме гаммирования. Определен в ГОСТ 34.13-2015.
    KuznyechikCtr,
    sys::KUZNYECHIK_CTR_SIZE,
    8,
    sys::KUZNYECHIK_CTR_ALIGNMENT
);
define_struct!(
    /// Блочный шифр с длиной блока 64 бит в режиме простой замены. Определен в ГОСТ 34.13-2015.
    MagmaEcb,
    sys::MAGMA_ECB_SIZE,
    4,
    sys::MAGMA_ECB_ALIGNMENT
);
define_struct!(
    /// Блочный шифр с длиной блока 128 бит в режиме простой замены. Определен в ГОСТ 34.13-2015.
    KuznyechikEcb,
    sys::KUZNYECHIK_ECB_SIZE,
    1,
    sys::KUZNYECHIK_ECB_ALIGNMENT
);
define_struct!(
    /// Функция хэширования с длиной хэш-кода 256 бит. Определена в ГОСТ Р 34.11-2012 (Стрибог).
    Streebog256,
    sys::STREEBOG256_SIZE,
    8,
    sys::STREEBOG256_ALIGNMENT
);
define_struct!(
    /// Функция хэширования с длиной хэш-кода 512 бит. Определена в ГОСТ Р 34.11-2012 (Стрибог).
    Streebog512,
    sys::STREEBOG512_SIZE,
    8,
    sys::STREEBOG512_ALIGNMENT
);
define_struct!(
    /// Режим выработки имитовставки с использованием шифра с длиной блока 64 бит. Определен в ГОСТ 34.13-2015.
    MagmaMac,
    sys::MAGMA_MAC_SIZE,
    8,
    sys::MAGMA_MAC_ALIGNMENT
);
define_struct!(
    /// Режим выработки имитовставки с использованием шифра с длиной блока 128 бит. Определен в ГОСТ 34.13-2015.
    KuznyechikMac,
    sys::KUZNYECHIK_MAC_SIZE,
    8,
    sys::KUZNYECHIK_MAC_ALIGNMENT
);
define_struct!(
    /// Потоковый шифр с длиной блока 64 бит в режиме гаммирования с обратной связью по выходу
    /// с размером синхропосылки 8 байт. Определен в ГОСТ Р 34.13-2015.
    MagmaOfb,
    sys::MAGMA_OFB_SIZE,
    4,
    sys::MAGMA_OFB_ALIGNMENT
);
define_struct!(
    /// Потоковый шифр с длиной блока 128 бит в режиме гаммирования с обратной связью по выходу
    /// с размером синхропосылки 16 байт. Определен в ГОСТ Р 34.13-2015.
    KuznyechikOfb,
    sys::KUZNYECHIK_OFB_SIZE,
    1,
    sys::KUZNYECHIK_OFB_ALIGNMENT
);

impl_drop!(KuznyechikMgm, sys::kuznyechik_mgm_drop);
impl_drop!(MagmaCbc, sys::magma_cbc_drop);
impl_drop!(KuznyechikCbc, sys::kuznyechik_cbc_drop);
impl_drop!(MagmaCfb, sys::magma_cfb_drop);
impl_drop!(KuznyechikCfb, sys::kuznyechik_cfb_drop);
impl_drop!(Magma, sys::magma_drop);
impl_drop!(Kuznyechik, sys::kuznyechik_drop);
impl_drop!(MagmaCtr, sys::magma_ctr_drop);
impl_drop!(KuznyechikCtr, sys::kuznyechik_ctr_drop);
impl_drop!(MagmaEcb, sys::magma_ecb_drop);
impl_drop!(KuznyechikEcb, sys::kuznyechik_ecb_drop);
impl_drop!(Streebog256, sys::streebog256_drop);
impl_drop!(Streebog512, sys::streebog512_drop);
impl_drop!(MagmaMac, sys::magma_mac_drop);
impl_drop!(KuznyechikMac, sys::kuznyechik_mac_drop);
impl_drop!(MagmaOfb, sys::magma_ofb_drop);
impl_drop!(KuznyechikOfb, sys::kuznyechik_ofb_drop);

/// Кривая `id-tc26-gost-3410-2012-256-paramSetA`, заданная в Р 1323565.1.024-2019.
pub struct Gost3410_2012_256A;
/// Кривая `id-tc26-gost-3410-2012-256-paramSetB`, заданная в Р 1323565.1.024-2019.
pub struct Gost3410_2012_256B;
/// Кривая `id-tc26-gost-3410-2012-256-paramSetC`, заданная в Р 1323565.1.024-2019.
pub struct Gost3410_2012_256C;
/// Кривая `id-tc26-gost-3410-2012-256-paramSetD`, заданная в Р 1323565.1.024-2019.
pub struct Gost3410_2012_256D;

impl private::Sealed for Gost3410_2012_256A {}
impl private::Sealed for Gost3410_2012_256B {}
impl private::Sealed for Gost3410_2012_256C {}
impl private::Sealed for Gost3410_2012_256D {}

impl Ec256Curve for Gost3410_2012_256A {
    const CURVE_ID: usize = sys::GOST3410_2012_256A as _;
}
impl Ec256Curve for Gost3410_2012_256B {
    const CURVE_ID: usize = sys::GOST3410_2012_256B as _;
}
impl Ec256Curve for Gost3410_2012_256C {
    const CURVE_ID: usize = sys::GOST3410_2012_256C as _;
}
impl Ec256Curve for Gost3410_2012_256D {
    const CURVE_ID: usize = sys::GOST3410_2012_256D as _;
}

/// Кривая `id-tc26-gost-3410-12-512-paramSetA`, заданная в Р 1323565.1.024-2019.
pub struct Gost3410_2012_512A;
/// Кривая `id-tc26-gost-3410-12-512-paramSetB`, заданная в Р 1323565.1.024-2019.
pub struct Gost3410_2012_512B;
/// Кривая `id-tc26-gost-3410-12-512-paramSetC`, заданная в Р 1323565.1.024-2019.
pub struct Gost3410_2012_512C;

impl private::Sealed for Gost3410_2012_512A {}
impl private::Sealed for Gost3410_2012_512B {}
impl private::Sealed for Gost3410_2012_512C {}

impl Ec512Curve for Gost3410_2012_512A {
    const CURVE_ID: usize = sys::GOST3410_2012_512A as _;
}
impl Ec512Curve for Gost3410_2012_512B {
    const CURVE_ID: usize = sys::GOST3410_2012_512B as _;
}
impl Ec512Curve for Gost3410_2012_512C {
    const CURVE_ID: usize = sys::GOST3410_2012_512C as _;
}

/// Подпись Ec256
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct Ec256Signature {
    inner: [u8; EC_256_DOUBLE_WIDTH],
}

impl Ec256Signature {
    /// Создать подпись
    pub fn new(bytes: [u8; EC_256_DOUBLE_WIDTH]) -> Self {
        Self { inner: bytes }
    }

    /// Создать подпись из среза
    pub fn from_slice(bytes: &[u8]) -> Option<Self> {
        if bytes.len() != EC_256_DOUBLE_WIDTH {
            return None;
        }

        let mut inner = [0; EC_256_DOUBLE_WIDTH];

        inner.copy_from_slice(bytes);

        Some(Self { inner })
    }

    /// Представить подпись как массив байт
    pub fn as_bytes(&self) -> [u8; EC_256_DOUBLE_WIDTH] {
        self.inner
    }

    fn as_ptr(&self) -> *const u8 {
        self.inner.as_ptr()
    }
}

/// Подпись Ec512
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct Ec512Signature {
    inner: [u8; EC_512_DOUBLE_WIDTH],
}

impl Ec512Signature {
    /// Создать подпись
    pub fn new(bytes: [u8; EC_512_DOUBLE_WIDTH]) -> Self {
        Self { inner: bytes }
    }

    /// Создать подпись из среза
    pub fn from_slice(bytes: &[u8]) -> Option<Self> {
        if bytes.len() != EC_512_DOUBLE_WIDTH {
            return None;
        }

        let mut inner = [0; EC_512_DOUBLE_WIDTH];

        inner.copy_from_slice(bytes);

        Some(Self { inner })
    }

    /// Представить подпись как массив байт
    pub fn as_bytes(&self) -> [u8; EC_512_DOUBLE_WIDTH] {
        self.inner
    }

    fn as_ptr(&self) -> *const u8 {
        self.inner.as_ptr()
    }
}

ec_struct!(
    /// Закрытый ключ Ec256
    Ec256PrivateKey,
    8,
    sys::EC_256_PRIVATE_KEY_ALIGNMENT,
    Ec256Curve
);
ec_struct!(
    /// Открытый ключ Ec256
    Ec256PublicKey,
    8,
    sys::EC_256_PUBLIC_KEY_ALIGNMENT,
    Ec256Curve
);
ec_struct!(
    /// Закрытый ключ Ec512
    Ec512PrivateKey,
    8,
    sys::EC_512_PRIVATE_KEY_ALIGNMENT,
    Ec512Curve
);
ec_struct!(
    /// Открытый ключ Ec512
    Ec512PublicKey,
    8,
    sys::EC_512_PUBLIC_KEY_ALIGNMENT,
    Ec512Curve
);

impl<C: Ec256Curve> Copy for Ec256PublicKey<C> {}
impl<C: Ec512Curve> Copy for Ec512PublicKey<C> {}

impl KuznyechikMgm {
    /// Создаёт контекст AEAD с использованием ключа `key`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::kuznyechik_mgm_new(cipher.as_mut_ptr(), key as *const _ as *mut _);

            cipher
        }
    }

    /// Зашифровывает сообщение на месте с присоединёнными данными и копирует соответствующую
    /// имитовставку в массив `tag`.
    ///
    /// - nonce — синхропосылка. Старший бит `nonce[0]` должен быть равен нулю.
    /// - aad — присоединённые данные
    /// - buf — подлежащие шифрованию данные
    /// - tag — имитовставка
    pub fn encrypt(
        &self,
        nonce: &[u8; KUZNYECHIK_MGM_NONCE_SIZE],
        aad: &[u8],
        buf: &mut [u8],
        tag: &mut [u8; KUZNYECHIK_MGM_TAG_SIZE],
    ) -> Result<(), ()> {
        unsafe {
            let res = sys::kuznyechik_mgm_encrypt(
                self as *const _ as *mut _,
                nonce as *const _ as *mut _,
                aad.as_ptr(),
                aad.len() as _,
                buf.as_mut_ptr(),
                buf.len() as _,
                tag.as_mut_ptr() as _,
            );

            if res == 0 {
                Ok(())
            } else {
                Err(())
            }
        }
    }

    /// Расшифровывает сообщение на месте с присоединёнными данными с использованием данной
    /// имитовставки `tag`. Возвращает ошибку в случае, если имитовставка не соответствует
    /// шифротексту.
    ///
    /// - nonce — синхропосылка
    /// - aad — присоединённые данные
    /// - buf — подлежащие расшифрованию данные
    /// - tag — имитовставка
    pub fn decrypt(
        &self,
        nonce: &[u8; KUZNYECHIK_MGM_NONCE_SIZE],
        aad: &[u8],
        buf: &mut [u8],
        tag: &mut [u8; KUZNYECHIK_MGM_TAG_SIZE],
    ) -> Result<(), ()> {
        unsafe {
            let res = sys::kuznyechik_mgm_decrypt(
                self as *const _ as *mut _,
                nonce as *const _ as *mut _,
                aad.as_ptr(),
                aad.len() as _,
                buf.as_mut_ptr(),
                buf.len() as _,
                tag.as_mut_ptr() as _,
            );

            if res == 0 {
                Ok(())
            } else {
                Err(())
            }
        }
    }
}

impl MagmaCbc {
    /// Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
    /// ключа `key` и синхропосылки `nonce`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE], nonce: &[u8; MAGMA_CBC_NONCE_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::magma_cbc_new(
                cipher.as_mut_ptr(),
                key as *const _ as *mut _,
                nonce as *const _ as *mut _,
            );

            cipher
        }
    }

    /// Шифрует блоки данных на месте.
    pub fn encrypt_blocks(&mut self, blocks: &mut [[u8; MAGMA_BLOCK_SIZE]]) {
        unsafe {
            sys::magma_cbc_encrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Расшифровывает блоки данных на месте.
    pub fn decrypt_blocks(&mut self, blocks: &mut [[u8; MAGMA_BLOCK_SIZE]]) {
        unsafe {
            sys::magma_cbc_decrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Шифрует срез как массив блоков
    pub fn encrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % MAGMA_BLOCK_SIZE {
            0 => slice.len() / MAGMA_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::magma_cbc_encrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }

    /// Расшифровывает срез как массив блоков
    pub fn decrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % MAGMA_BLOCK_SIZE {
            0 => slice.len() / MAGMA_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::magma_cbc_decrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }

    /// Зашифровывает сообщение на месте, поглощая контекст.
    ///
    /// Байты `0..pos` буфера `buf` используются как сообщение, когда же байты `pos..buf_len`
    /// используются как зарезервированное место для дополнения.
    pub fn encrypt_pad(mut self, buf: &mut [u8], pos: usize) -> Result<usize, ()> {
        unsafe {
            let res = sys::magma_cbc_encrypt_pad(
                self.as_mut_ptr(),
                buf.as_mut_ptr(),
                buf.len() as _,
                pos as _,
            );

            if res >= 0 {
                Ok(res as usize)
            } else {
                Err(())
            }
        }
    }

    /// Расшифровывает сообщение на месте, поглощая контекст шифра.
    ///
    /// Возвращает ошибку в случае, когда длина буфера не делится на [MAGMA_BLOCK_SIZE] или
    /// расшифрованное сообщение имеет некорректное дополнение.
    pub fn decrypt_pad(mut self, buf: &mut [u8]) -> Result<usize, ()> {
        unsafe {
            let res =
                sys::magma_cbc_decrypt_pad(self.as_mut_ptr(), buf.as_mut_ptr(), buf.len() as _);

            if res >= 0 {
                Ok(res as usize)
            } else {
                Err(())
            }
        }
    }
}

impl KuznyechikCbc {
    /// Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
    /// ключа `key` и синхропосылки `nonce`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE], nonce: &[u8; KUZNYECHIK_CBC_NONCE_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::kuznyechik_cbc_new(
                cipher.as_mut_ptr(),
                key.as_ptr() as _,
                nonce as *const _ as *mut _,
            );

            cipher
        }
    }

    /// Шифрует блоки данных на месте.
    pub fn encrypt_blocks(&mut self, blocks: &mut [[u8; KUZNYECHIK_BLOCK_SIZE]]) {
        unsafe {
            sys::kuznyechik_cbc_encrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Расшифровывает блоки данных на месте.
    pub fn decrypt_blocks(&mut self, blocks: &mut [[u8; KUZNYECHIK_BLOCK_SIZE]]) {
        unsafe {
            sys::kuznyechik_cbc_decrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Шифрует срез как массив блоков
    pub fn encrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % KUZNYECHIK_BLOCK_SIZE {
            0 => slice.len() / KUZNYECHIK_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::kuznyechik_cbc_encrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }

    /// Расшифровывает срез как массив блоков
    pub fn decrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % KUZNYECHIK_BLOCK_SIZE {
            0 => slice.len() / KUZNYECHIK_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::kuznyechik_cbc_decrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }

    /// Зашифровывает сообщение на месте, поглощая контекст.
    ///
    /// Байты `0..pos` буфера `buf` используются как сообщение, когда же байты `pos..buf_len`
    /// используются как зарезервированное место для дополнения.
    pub fn encrypt_pad(mut self, buf: &mut [u8], pos: usize) -> Result<usize, ()> {
        unsafe {
            let res = sys::kuznyechik_cbc_encrypt_pad(
                self.as_mut_ptr(),
                buf.as_mut_ptr(),
                buf.len() as _,
                pos as _,
            );

            if res >= 0 {
                Ok(res as usize)
            } else {
                Err(())
            }
        }
    }

    /// Расшифровывает сообщение на месте, поглощая контекст шифра.
    ///
    /// Возвращает ошибку в случае, когда длина буфера не делится на [MAGMA_BLOCK_SIZE] или
    /// расшифрованное сообщение имеет некорректное дополнение.
    pub fn decrypt_pad(mut self, buf: &mut [u8]) -> Result<usize, ()> {
        unsafe {
            let res = sys::kuznyechik_cbc_decrypt_pad(
                self.as_mut_ptr(),
                buf.as_mut_ptr(),
                buf.len() as _,
            );

            if res >= 0 {
                Ok(res as usize)
            } else {
                Err(())
            }
        }
    }
}

impl MagmaCfb {
    /// Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по шифртексту
    /// с использованием ключа `key` и синхропосылки `nonce`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE], nonce: &[u8; MAGMA_CFB_NONCE_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::magma_cfb_new(
                cipher.as_mut_ptr(),
                key as *const _ as *mut _,
                nonce as *const _ as *mut _,
            );

            cipher
        }
    }

    /// Зашифровывает сообщение на месте.
    pub fn encrypt(&mut self, buf: &mut [u8]) {
        unsafe { sys::magma_cfb_encrypt(self.as_mut_ptr(), buf.as_mut_ptr(), buf.len() as _) }
    }

    /// Расшифровывает сообщение на месте.
    pub fn decrypt(&mut self, buf: &mut [u8]) {
        unsafe { sys::magma_cfb_decrypt(self.as_mut_ptr(), buf.as_mut_ptr(), buf.len() as _) }
    }
}

impl KuznyechikCfb {
    /// Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по шифртексту
    /// с использованием ключа `key` и синхропосылки `nonce`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE], nonce: &[u8; KUZNYECHIK_CFB_NONCE_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::kuznyechik_cfb_new(
                cipher.as_mut_ptr(),
                key as *const _ as *mut _,
                nonce as *const _ as *mut _,
            );

            cipher
        }
    }

    /// Зашифровывает сообщение на месте.
    pub fn encrypt(&mut self, buf: &mut [u8]) {
        unsafe { sys::kuznyechik_cfb_encrypt(self.as_mut_ptr(), buf.as_mut_ptr(), buf.len() as _) }
    }

    /// Расшифровывает сообщение на месте.
    pub fn decrypt(&mut self, buf: &mut [u8]) {
        unsafe { sys::kuznyechik_cfb_decrypt(self.as_mut_ptr(), buf.as_mut_ptr(), buf.len() as _) }
    }
}

impl Magma {
    /// Создает новый контекст блочного шифра.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::magma_new(cipher.as_mut_ptr(), key as *const _ as *mut _);

            cipher
        }
    }

    /// Шифрует блок на месте.
    pub fn encrypt_block(&mut self, block: &mut [u8; MAGMA_BLOCK_SIZE]) {
        unsafe { sys::magma_encrypt_block(self.as_mut_ptr(), block.as_mut_ptr() as _) }
    }

    /// Расшифровывает блок на месте.
    pub fn decrypt_block(&mut self, block: &mut [u8; MAGMA_BLOCK_SIZE]) {
        unsafe { sys::magma_decrypt_block(self.as_mut_ptr(), block.as_mut_ptr() as _) }
    }

    /// Шифрует массив блоков на месте.
    pub fn encrypt_blocks(&mut self, blocks: &mut [[u8; MAGMA_BLOCK_SIZE]]) {
        unsafe {
            sys::magma_encrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Расшифровывает массив блоков на месте.
    pub fn decrypt_blocks(&mut self, blocks: &mut [[u8; MAGMA_BLOCK_SIZE]]) {
        unsafe {
            sys::magma_decrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Шифрует срез как массив блоков
    pub fn encrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % MAGMA_BLOCK_SIZE {
            0 => slice.len() / MAGMA_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::magma_encrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }

    /// Расшифровывает срез как массив блоков
    pub fn decrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % MAGMA_BLOCK_SIZE {
            0 => slice.len() / MAGMA_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::magma_decrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }
}

impl Kuznyechik {
    /// Создает новый контекст блочного шифра.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::kuznyechik_new(cipher.as_mut_ptr(), key.as_ptr() as _);

            cipher
        }
    }

    /// Шифрует блок на месте.
    pub fn encrypt_block(&mut self, block: &mut [u8; KUZNYECHIK_BLOCK_SIZE]) {
        unsafe { sys::kuznyechik_encrypt_block(self.as_mut_ptr(), block.as_mut_ptr() as _) }
    }

    /// Расшифровывает блок на месте.
    pub fn decrypt_block(&mut self, block: &mut [u8; KUZNYECHIK_BLOCK_SIZE]) {
        unsafe { sys::kuznyechik_decrypt_block(self.as_mut_ptr(), block.as_mut_ptr() as _) }
    }

    /// Шифрует массив блоков на месте.
    pub fn encrypt_blocks(&mut self, blocks: &mut [[u8; KUZNYECHIK_BLOCK_SIZE]]) {
        unsafe {
            sys::kuznyechik_encrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Расшифровывает массив блоков на месте.
    pub fn decrypt_blocks(&mut self, blocks: &mut [[u8; KUZNYECHIK_BLOCK_SIZE]]) {
        unsafe {
            sys::kuznyechik_decrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Шифрует срез как массив блоков
    pub fn encrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % KUZNYECHIK_BLOCK_SIZE {
            0 => slice.len() / KUZNYECHIK_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::kuznyechik_encrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }

    /// Расшифровывает срез как массив блоков
    pub fn decrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % KUZNYECHIK_BLOCK_SIZE {
            0 => slice.len() / KUZNYECHIK_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::kuznyechik_decrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }
}

impl MagmaCtr {
    /// Создаёт контекст потокового шифра в режиме гаммирования с использованием ключа `key`
    /// и синхропосылки `nonce`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE], nonce: &[u8; MAGMA_CTR_NONCE_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::magma_ctr_new(
                cipher.as_mut_ptr(),
                key as *const _ as *mut _,
                nonce as *const _ as *mut _,
            );

            cipher
        }
    }

    /// Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
    /// шифра.
    ///
    /// Возвращает ошибку в случае, когда достигнут конец ключевого потока.
    pub fn apply_keystream(&mut self, buf: &mut [u8]) -> Result<(), ()> {
        let res = unsafe {
            sys::magma_ctr_apply_keystream(self.as_mut_ptr(), buf.as_mut_ptr(), buf.len() as _)
        };
        match res {
            0 => Ok(()),
            _ => Err(()),
        }
    }

    /// Устанавливает текующую позицию ключевого потока шифра.
    ///
    /// Возвращает ошибку в случае, когда требуемая позиция превышает длину ключевого потока.
    pub fn seek(&mut self, pos: u64) -> Result<(), ()> {
        let res = unsafe { sys::magma_ctr_seek(self.as_mut_ptr(), pos) };
        match res {
            0 => Ok(()),
            _ => Err(()),
        }
    }

    /// Возвращает текущую позицию ключевого потока шифра.
    pub fn current_pos(&self) -> u64 {
        unsafe { sys::magma_ctr_get_pos(self as *const _ as *mut _) }
    }
}

impl KuznyechikCtr {
    /// Создаёт контекст потокового шифра в режиме гаммирования с использованием ключа `key`
    /// и синхропосылки `nonce`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE], nonce: &[u8; KUZNYECHIK_CTR_NONCE_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::kuznyechik_ctr_new(
                cipher.as_mut_ptr(),
                key as *const _ as *mut _,
                nonce as *const _ as *mut _,
            );

            cipher
        }
    }

    /// Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
    /// шифра.
    ///
    /// Возвращает ошибку в случае, когда достигнут конец ключевого потока.
    pub fn apply_keystream(&mut self, buf: &mut [u8]) -> Result<(), ()> {
        let res = unsafe {
            sys::kuznyechik_ctr_apply_keystream(self.as_mut_ptr(), buf.as_mut_ptr(), buf.len() as _)
        };

        match res {
            0 => Ok(()),
            _ => Err(()),
        }
    }

    /// Устанавливает текующую позицию ключевого потока шифра.
    ///
    /// Возвращает ошибку в случае, когда требуемая позиция превышает длину ключевого потока.
    pub fn seek(&mut self, pos: u64) -> Result<(), ()> {
        let res = unsafe { sys::kuznyechik_ctr_seek(self.as_mut_ptr(), pos) };

        match res {
            0 => Ok(()),
            _ => Err(()),
        }
    }

    /// Возвращает текущую позицию ключевого потока шифра.
    pub fn current_pos(&self) -> u64 {
        unsafe { sys::kuznyechik_ctr_get_pos(self as *const _ as *mut _) }
    }
}

impl MagmaEcb {
    /// Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
    /// ключа `key`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::magma_ecb_new(cipher.as_mut_ptr(), key as *const _ as *mut _);

            cipher
        }
    }

    /// Шифрует блоки на месте.
    pub fn encrypt_blocks(&mut self, blocks: &mut [[u8; MAGMA_BLOCK_SIZE]]) {
        unsafe {
            sys::magma_ecb_encrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Расшифровывает блоки на месте.
    pub fn decrypt_blocks(&mut self, blocks: &mut [[u8; MAGMA_BLOCK_SIZE]]) {
        unsafe {
            sys::magma_ecb_decrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Шифрует срез как массив блоков
    pub fn encrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % MAGMA_BLOCK_SIZE {
            0 => slice.len() / MAGMA_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::magma_ecb_encrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }

    /// Расшифровывает срез как массив блоков
    pub fn decrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % MAGMA_BLOCK_SIZE {
            0 => slice.len() / MAGMA_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::magma_ecb_decrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }

    /// Зашифровывает сообщение на месте, поглощая контекст шифра. Возвращает длину шифротекста.
    ///
    /// Байты `0..pos` буфера `buf` используются как сообщение, когда же байты `pos..buf_len`
    /// используются как зарезервированное место для дополнения.
    ///
    /// Возвращает ошибку в случае, когда места для дополнения недостаточно.
    /// велико.
    pub fn encrypt_pad(mut self, buf: &mut [u8], pos: usize) -> Result<usize, ()> {
        unsafe {
            let res = sys::magma_ecb_encrypt_pad(
                self.as_mut_ptr(),
                buf.as_mut_ptr(),
                buf.len() as _,
                pos as _,
            );

            if res >= 0 {
                Ok(res as usize)
            } else {
                Err(())
            }
        }
    }

    /// Расшифровывает сообщение на месте, поглощая контекст шифра. Возвращает длину сообщения.
    ///
    /// Возвращает ошибку в случае, когда длина буфера не делится на #MAGMA_BLOCK_SIZE или
    /// расшифрованное сообщение имеет некорректное дополнение.
    pub fn decrypt_pad(mut self, buf: &mut [u8]) -> Result<usize, ()> {
        unsafe {
            let res =
                sys::magma_ecb_decrypt_pad(self.as_mut_ptr(), buf.as_mut_ptr(), buf.len() as _);

            if res >= 0 {
                Ok(res as usize)
            } else {
                Err(())
            }
        }
    }
}

impl KuznyechikEcb {
    /// Создаёт контекст блочного шифра в режиме простой замены с зацеплением с использованием
    /// ключа `key`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::kuznyechik_ecb_new(cipher.as_mut_ptr(), key as *const _ as *mut _);

            cipher
        }
    }

    /// Шифрует блоки на месте.
    pub fn encrypt_blocks(&mut self, blocks: &mut [[u8; KUZNYECHIK_BLOCK_SIZE]]) {
        unsafe {
            sys::kuznyechik_ecb_encrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Расшифровывает блоки на месте.
    pub fn decrypt_blocks(&mut self, blocks: &mut [[u8; KUZNYECHIK_BLOCK_SIZE]]) {
        unsafe {
            sys::kuznyechik_ecb_decrypt_blocks(
                self.as_mut_ptr(),
                blocks.as_mut_ptr() as _,
                blocks.len() as _,
            );
        }
    }

    /// Шифрует срез как массив блоков
    pub fn encrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % KUZNYECHIK_BLOCK_SIZE {
            0 => slice.len() / KUZNYECHIK_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::kuznyechik_ecb_encrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }

    /// Расшифровывает срез как массив блоков
    pub fn decrypt_slice(&mut self, slice: &mut [u8]) -> Result<(), ()> {
        let len = match slice.len() % KUZNYECHIK_BLOCK_SIZE {
            0 => slice.len() / KUZNYECHIK_BLOCK_SIZE,
            _ => return Err(()),
        };

        unsafe {
            sys::kuznyechik_ecb_decrypt_blocks(self.as_mut_ptr(), slice.as_mut_ptr() as _, len);
        }

        Ok(())
    }

    /// Зашифровывает сообщение на месте, поглощая контекст шифра. Возвращает длину шифротекста.
    ///
    /// Байты `0..pos` буфера `buf` используются как сообщение, когда же байты `pos..buf_len`
    /// используются как зарезервированное место для дополнения.
    ///
    /// Возвращает ошибку в случае, когда места для дополнения недостаточно.
    /// велико.
    pub fn encrypt_pad(mut self, buf: &mut [u8], pos: usize) -> Result<usize, ()> {
        unsafe {
            let res = sys::kuznyechik_ecb_encrypt_pad(
                self.as_mut_ptr(),
                buf.as_mut_ptr(),
                buf.len() as _,
                pos as _,
            );

            if res >= 0 {
                Ok(res as usize)
            } else {
                Err(())
            }
        }
    }

    /// Расшифровывает сообщение на месте, поглощая контекст шифра. Возвращает длину сообщения.
    ///
    /// Возвращает ошибку в случае, когда длина буфера не делится на #MAGMA_BLOCK_SIZE или
    /// расшифрованное сообщение имеет некорректное дополнение.
    pub fn decrypt_pad(mut self, buf: &mut [u8]) -> Result<usize, ()> {
        unsafe {
            let res = sys::kuznyechik_ecb_decrypt_pad(
                self.as_mut_ptr(),
                buf.as_mut_ptr(),
                buf.len() as _,
            );

            if res >= 0 {
                Ok(res as usize)
            } else {
                Err(())
            }
        }
    }
}

/// Трейт хэш-функции Стрибог
pub trait StreebogHasher {
    /// Конкретный вариант хэш-функции
    type Hash;

    /// Создать состояние функции хэширования
    fn new() -> Self;
    /// хэшировать сообщение
    fn digest(msg: &[u8]) -> Self::Hash;
    /// Обновить состояние функции хэширования
    fn update(&mut self, msg: &[u8]);
    /// Финализировать состояние функции хэширования
    fn finalize(self) -> Self::Hash;
    /// Финализировать состояние функции хэширования и заменить его новым
    fn finalize_reset(&mut self) -> Self::Hash;
    /// Сбросить состояние функции хэширования
    fn reset(&mut self);
}

impl StreebogHasher for Streebog256 {
    type Hash = [u8; STREEBOG256_RESULT_SIZE];

    fn new() -> Self {
        unsafe {
            let mut hasher = MaybeUninit::<Self>::uninit();

            sys::streebog256_new(hasher.as_mut_ptr().cast());

            hasher.assume_init()
        }
    }

    fn digest(msg: &[u8]) -> Self::Hash {
        unsafe {
            let mut res = MaybeUninit::<Self::Hash>::uninit();
            sys::streebog256_digest(msg.as_ptr(), msg.len() as _, res.as_mut_ptr().cast());
            res.assume_init()
        }
    }

    fn update(&mut self, msg: &[u8]) {
        unsafe { sys::streebog256_update(self.as_mut_ptr(), msg.as_ptr(), msg.len() as _) }
    }

    fn finalize(mut self) -> Self::Hash {
        unsafe {
            let mut res = MaybeUninit::<Self::Hash>::uninit();
            sys::streebog256_finalize(self.as_mut_ptr(), res.as_mut_ptr().cast());
            res.assume_init()
        }
    }

    fn finalize_reset(&mut self) -> Self::Hash {
        unsafe {
            let mut res = MaybeUninit::<Self::Hash>::uninit();
            sys::streebog256_finalize_reset(self.as_mut_ptr(), res.as_mut_ptr().cast());
            res.assume_init()
        }
    }

    fn reset(&mut self) {
        unsafe { sys::streebog256_reset(self.as_mut_ptr()) }
    }
}

impl Default for Streebog256 {
    fn default() -> Self {
        Self::new()
    }
}

impl StreebogHasher for Streebog512 {
    type Hash = [u8; STREEBOG512_RESULT_SIZE];

    fn new() -> Self {
        unsafe {
            let mut hasher = MaybeUninit::<Self>::uninit();

            sys::streebog512_new(hasher.as_mut_ptr().cast());

            hasher.assume_init()
        }
    }

    fn digest(msg: &[u8]) -> Self::Hash {
        unsafe {
            let mut res = MaybeUninit::<Self::Hash>::uninit();
            sys::streebog512_digest(msg.as_ptr(), msg.len() as _, res.as_mut_ptr().cast());
            res.assume_init()
        }
    }

    fn update(&mut self, msg: &[u8]) {
        unsafe { sys::streebog512_update(self.as_mut_ptr(), msg.as_ptr(), msg.len() as _) }
    }

    fn finalize(mut self) -> Self::Hash {
        unsafe {
            let mut res = MaybeUninit::<Self::Hash>::uninit();
            sys::streebog512_finalize(self.as_mut_ptr(), res.as_mut_ptr().cast());
            res.assume_init()
        }
    }

    fn finalize_reset(&mut self) -> Self::Hash {
        unsafe {
            let mut res = MaybeUninit::<Self::Hash>::uninit();
            sys::streebog512_finalize_reset(self.as_mut_ptr(), res.as_mut_ptr().cast());
            res.assume_init()
        }
    }

    fn reset(&mut self) {
        unsafe { sys::streebog512_reset(self.as_mut_ptr()) }
    }
}

impl Default for Streebog512 {
    fn default() -> Self {
        Self::new()
    }
}

impl MagmaMac {
    /// Создаёт новый состояние режима выработки имитовставки с использованием ключа `key`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::magma_mac_new(cipher.as_mut_ptr(), key.as_ptr() as _);

            cipher
        }
    }

    /// Скармливает сообщение режиму выработки имитовставки, обновляя его состояние.
    pub fn update(&mut self, msg: &[u8]) {
        unsafe { sys::magma_mac_update(self.as_mut_ptr(), msg.as_ptr(), msg.len() as _) }
    }

    /// Финализирует состояние режима выработки имитовставки и возвращает результирующую имитовставку.
    /// Поглощает состояние.
    pub fn finalize(mut self) -> [u8; MAGMA_MAC_RESULT_SIZE] {
        unsafe {
            let mut res = [0; MAGMA_MAC_RESULT_SIZE];

            sys::magma_mac_finalize(self.as_mut_ptr(), res.as_mut_ptr() as _);

            res
        }
    }

    /// Финализирует состояние режима выработки имитовставки и возвращает результирующую имитовставку.
    /// Заменяет существующее состояние на новое.
    pub fn finalize_reset(&mut self) -> [u8; MAGMA_MAC_RESULT_SIZE] {
        unsafe {
            let mut res = [0; MAGMA_MAC_RESULT_SIZE];

            sys::magma_mac_finalize_reset(self.as_mut_ptr(), res.as_mut_ptr() as _);

            res
        }
    }

    /// Проверяет соответствие данных, скормленных режиму выработки имитовставки, данной имитовставке.
    ///
    /// Возвращает `Ok(())` в случае соответствия и `Err(())` в случае несоответствия.
    pub fn verify(mut self, tag: &[u8]) -> Result<(), ()> {
        unsafe {
            let res = sys::magma_mac_verify(self.as_mut_ptr(), tag as *const _ as *mut _);

            if res {
                Ok(())
            } else {
                Err(())
            }
        }
    }

    /// Заменяет состояние режима выработки имитовставки новым состоянием.
    pub fn reset(&mut self) {
        unsafe { sys::magma_mac_reset(self.as_mut_ptr()) }
    }
}

impl KuznyechikMac {
    /// Создаёт новый состояние режима выработки имитовставки с использованием ключа `key`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::kuznyechik_mac_new(cipher.as_mut_ptr(), key as *const _ as *mut _);

            cipher
        }
    }

    /// Скармливает сообщение режиму выработки имитовставки, обновляя его состояние.
    pub fn update(&mut self, msg: &[u8]) {
        unsafe { sys::kuznyechik_mac_update(self.as_mut_ptr(), msg.as_ptr(), msg.len() as _) }
    }

    /// Финализирует состояние режима выработки имитовставки и возвращает результирующую имитовставку.
    /// Поглощает состояние.
    pub fn finalize(mut self) -> [u8; KUZNYECHIK_MAC_RESULT_SIZE] {
        unsafe {
            let mut res = [0; KUZNYECHIK_MAC_RESULT_SIZE];

            sys::kuznyechik_mac_finalize(self.as_mut_ptr(), res.as_mut_ptr() as _);

            res
        }
    }

    /// Финализирует состояние режима выработки имитовставки и возвращает результирующую имитовставку.
    /// Заменяет существующее состояние на новое.
    pub fn finalize_reset(&mut self) -> [u8; KUZNYECHIK_MAC_RESULT_SIZE] {
        unsafe {
            let mut res = [0; KUZNYECHIK_MAC_RESULT_SIZE];

            sys::kuznyechik_mac_finalize_reset(self.as_mut_ptr(), res.as_mut_ptr() as _);

            res
        }
    }

    /// Проверяет соответствие данных, скормленных режиму выработки имитовставки, данной имитовставке.
    ///
    /// Возвращает `Ok(())` в случае соответствия и `Err(())` в случае несоответствия.
    pub fn verify(mut self, tag: &[u8]) -> Result<(), ()> {
        unsafe {
            let res = sys::kuznyechik_mac_verify(self.as_mut_ptr(), tag as *const _ as *mut _);

            if res {
                Ok(())
            } else {
                Err(())
            }
        }
    }

    /// Заменяет состояние режима выработки имитовставки новым состоянием.
    pub fn reset(&mut self) {
        unsafe { sys::kuznyechik_mac_reset(self.as_mut_ptr()) }
    }
}

impl MagmaOfb {
    /// Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по выходу
    /// с использованием ключа `key` и синхропосылки `nonce`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE], nonce: &[u8; MAGMA_OFB_NONCE_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::magma_ofb_new(
                cipher.as_mut_ptr(),
                key as *const _ as *mut _,
                nonce as *const _ as *mut _,
            );

            cipher
        }
    }

    /// Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
    /// шифра.
    pub fn apply_keystream(&mut self, buf: &mut [u8]) {
        unsafe {
            sys::magma_ofb_apply_keystream(self.as_mut_ptr(), buf.as_mut_ptr(), buf.len() as _);
        }
    }
}

impl KuznyechikOfb {
    /// Создаёт контекст потокового шифра в режиме гаммирования с обратной связью по выходу
    /// с использованием ключа `key` и синхропосылки `nonce`.
    pub fn new(key: &[u8; SYMMETRIC_KEY_SIZE], nonce: &[u8; KUZNYECHIK_OFB_NONCE_SIZE]) -> Self {
        unsafe {
            let mut cipher: Self = zeroed();

            sys::kuznyechik_ofb_new(
                cipher.as_mut_ptr(),
                key as *const _ as *mut _,
                nonce as *const _ as *mut _,
            );

            cipher
        }
    }

    /// Шифрует или расшифровывает сообщение на месте посредством гаммирования его ключевым потоком
    /// шифра.
    pub fn apply_keystream(&mut self, buf: &mut [u8]) {
        unsafe {
            sys::kuznyechik_ofb_apply_keystream(
                self.as_mut_ptr(),
                buf.as_mut_ptr(),
                buf.len() as _,
            );
        }
    }
}

impl<C: Ec256Curve> Ec256PrivateKey<C> {
    /// Создаёт закрытый ключ
    pub fn generate<R: Rng + CryptoRng>(rng: &mut R) -> Self {
        unsafe {
            let cid = C::CURVE_ID;
            let mut seed = [0; EC_256_WIDTH];
            let mut private_key = Self::zeroed();

            loop {
                rng.fill_bytes(&mut seed);

                let res =
                    sys::ec256_gen_private_key(cid, private_key.as_mut_ptr(), seed.as_mut_ptr());

                if res == 0 {
                    break;
                }
            }

            private_key
        }
    }

    /// Получает открытый ключ из закрытого ключа
    pub fn public_key(&self) -> Ec256PublicKey<C> {
        unsafe {
            let cid = C::CURVE_ID;
            let mut public_key = Ec256PublicKey::zeroed();

            // Ignore error because type system prevents us from this error
            let _res =
                sys::ec256_get_public_key(cid, self.as_ptr() as *mut _, public_key.as_mut_ptr());

            public_key
        }
    }

    /// Десериализует закрытый ключ из представления ГОСТ 34.10-2012 раздел 5.2
    pub fn from_bytes(bytes: &[u8]) -> Result<Self, ()> {
        unsafe {
            if bytes.len() != EC_256_WIDTH {
                return Err(());
            }

            let cid = C::CURVE_ID;
            let mut private_key = Self::zeroed();

            let res =
                sys::ec256_private_key_from_bytes(cid, bytes.as_ptr(), private_key.as_mut_ptr());

            if res != 0 {
                return Err(());
            }

            Ok(private_key)
        }
    }

    /// Сериализует закрытый ключ в представление ГОСТ 34.10-2012 раздел 5.2
    pub fn to_bytes(&self) -> [u8; EC_256_WIDTH] {
        unsafe {
            let cid = C::CURVE_ID;
            let mut res = [0; EC_256_WIDTH];

            // Ignore error because type system prevents us from this error
            let _res =
                sys::ec256_private_key_to_bytes(cid, self.as_ptr() as *mut _, res.as_mut_ptr());

            res
        }
    }

    /// Подписывает хэш сообщения с помощью закрытого ключа. По ГОСТу хэш сообщения должен быть
    /// вычислен с помощью функции Стрибог.
    pub fn sign_hash<R: Rng + CryptoRng>(
        &self,
        hash: &[u8],
        rng: &mut R,
    ) -> Result<Ec256Signature, ()> {
        unsafe {
            if hash.len() != EC_256_WIDTH {
                return Err(());
            }

            let cid = C::CURVE_ID;
            let mut seed = [0; EC_256_WIDTH];
            let mut inner = [0; EC_256_DOUBLE_WIDTH];

            loop {
                rng.fill_bytes(&mut seed);

                let res = sys::ec256_dsa_sign_hash(
                    cid,
                    self.as_ptr() as *mut _,
                    hash.as_ptr(),
                    seed.as_ptr(),
                    inner.as_mut_ptr(),
                );

                if res == 0 {
                    break;
                }
            }

            Ok(Ec256Signature { inner })
        }
    }

    /// Подписывает сообщение с помощью закрытого ключа
    pub fn sign<R: Rng + CryptoRng>(&self, msg: &[u8], rng: &mut R) -> Ec256Signature {
        unsafe {
            let cid = C::CURVE_ID;
            let mut seed = [0; EC_256_WIDTH];
            let mut inner = [0; EC_256_DOUBLE_WIDTH];

            loop {
                rng.fill_bytes(&mut seed);

                let res = sys::ec256_dsa_sign(
                    cid,
                    self.as_ptr() as *mut _,
                    msg.as_ptr(),
                    msg.len(),
                    seed.as_ptr(),
                    inner.as_mut_ptr(),
                );

                if res == 0 {
                    break;
                }
            }

            Ec256Signature { inner }
        }
    }

    /// Вырабатывает алгоритм выработки ключей обмена длиной 512 бит, используя 512-битные
    /// эллиптические кривые.
    ///
    /// Если `ukm` равен `None`, то используется значение по умолчанию равное единице. Если же
    /// `ukm` не `None`, то значение внутри `Some` не должно быть равным нулю по модулю простого
    /// числа эллиптической кривой. В случае передачи такого значения возвращается ошибка. Если
    /// используемый приватный или публичный ключ статичны (т.е. используются неэфемерные ключи),
    /// то рекомендуется использовать `ukm`, содержащий не менее 128 переменных бит.
    pub fn vko256(
        &self,
        other_pk: &Ec256PublicKey<C>,
        ukm: Option<[u8; EC_256_WIDTH]>,
    ) -> Result<[u8; EC_256_WIDTH], ()> {
        let cid = C::CURVE_ID;
        let mut out = MaybeUninit::<[u8; EC_256_WIDTH]>::uninit();

        let ukm = match ukm {
            None => core::ptr::null(),
            Some(ref v) => v as *const _,
        };

        unsafe {
            let res = sys::ec256_vko256(
                cid,
                self.as_ptr() as *mut _,
                other_pk.as_ptr() as *mut _,
                ukm,
                out.as_mut_ptr().cast(),
            );

            if res == 0 {
                Ok(out.assume_init())
            } else {
                Err(())
            }
        }
    }
}

impl<C: Ec256Curve> Drop for Ec256PrivateKey<C> {
    fn drop(&mut self) {
        unsafe {
            sys::ec256_private_key_drop(self.as_mut_ptr());
        }
    }
}

impl<C: Ec256Curve> Ec256PublicKey<C> {
    /// Десериализует открытый ключ из представления ГОСТ 34.10-2012 раздел 5.2
    pub fn from_bytes(bytes: &[u8]) -> Result<Self, ()> {
        unsafe {
            if bytes.len() != EC_256_DOUBLE_WIDTH {
                return Err(());
            }

            let cid = C::CURVE_ID;
            let mut public_key = Self::zeroed();

            let res =
                sys::ec256_public_key_from_bytes(cid, bytes.as_ptr(), public_key.as_mut_ptr());

            if res != 0 {
                return Err(());
            }

            Ok(public_key)
        }
    }

    /// Сериализует открытый ключ из внутреннего представления в представление ГОСТ 34.10-2012 раздел 5.2
    pub fn to_bytes(&self) -> [u8; EC_256_DOUBLE_WIDTH] {
        unsafe {
            let cid = C::CURVE_ID;
            let mut res = [0; EC_256_DOUBLE_WIDTH];

            // Ignore error because type system prevents us from this error
            let _res =
                sys::ec256_public_key_to_bytes(cid, self.as_ptr() as *mut _, res.as_mut_ptr());

            res
        }
    }

    /// Проверяет подпись хэша сообщения с помощью публичной ключа. По ГОСТу хэш сообщения должен
    /// быть вычислен с помощью функции Стрибог.
    pub fn verify_hash(&self, hash: &[u8], signature: &Ec256Signature) -> Result<(), ()> {
        unsafe {
            let cid = C::CURVE_ID;

            let res = sys::ec256_dsa_verify_hash(
                cid,
                self.as_ptr() as *mut _,
                hash.as_ptr(),
                signature.as_ptr(),
            );

            if res != 0 {
                return Err(());
            }

            Ok(())
        }
    }

    /// Проверяет подпись сообщения с помощью публичной ключа
    pub fn verify(&self, msg: &[u8], signature: &Ec256Signature) -> Result<(), ()> {
        unsafe {
            let cid = C::CURVE_ID;

            let res = sys::ec256_dsa_verify(
                cid,
                self.as_ptr() as *mut _,
                msg.as_ptr(),
                msg.len(),
                signature.as_ptr(),
            );

            if res != 0 {
                return Err(());
            }

            Ok(())
        }
    }
}

impl<C: Ec512Curve> Ec512PrivateKey<C> {
    /// Создаёт закрытый ключ
    pub fn generate<R: Rng + CryptoRng>(rng: &mut R) -> Self {
        unsafe {
            let cid = C::CURVE_ID;
            let mut seed = [0; EC_512_WIDTH];
            let mut private_key = Self::zeroed();

            loop {
                rng.fill_bytes(&mut seed);

                let res =
                    sys::ec512_gen_private_key(cid, private_key.as_mut_ptr(), seed.as_mut_ptr());

                if res == 0 {
                    break;
                }
            }

            private_key
        }
    }

    /// Получает открытый ключ из закрытого ключа
    pub fn public_key(&self) -> Ec512PublicKey<C> {
        unsafe {
            let cid = C::CURVE_ID;
            let mut public_key = Ec512PublicKey::zeroed();

            // Ignore error because type system prevents us from this error
            let _res =
                sys::ec512_get_public_key(cid, self.as_ptr() as *mut _, public_key.as_mut_ptr());

            public_key
        }
    }

    /// Десериализует закрытый ключ из представления ГОСТ 34.10-2012 раздел 5.2
    pub fn from_bytes(bytes: &[u8]) -> Result<Self, ()> {
        unsafe {
            if bytes.len() != EC_512_WIDTH {
                return Err(());
            }

            let cid = C::CURVE_ID;
            let mut private_key = Self::zeroed();

            let res =
                sys::ec512_private_key_from_bytes(cid, bytes.as_ptr(), private_key.as_mut_ptr());

            if res != 0 {
                return Err(());
            }

            Ok(private_key)
        }
    }

    /// Сериализует закрытый ключ в представление ГОСТ 34.10-2012 раздел 5.2
    pub fn to_bytes(&self) -> [u8; EC_512_WIDTH] {
        unsafe {
            let cid = C::CURVE_ID;
            let mut res = [0; EC_512_WIDTH];

            // Ignore error because type system prevents us from this error
            let _res =
                sys::ec512_private_key_to_bytes(cid, self.as_ptr() as *mut _, res.as_mut_ptr());

            res
        }
    }

    /// Подписывает хэш сообщения с помощью закрытого ключа. По ГОСТу хэш сообщения должен быть
    /// вычислен с помощью функции Стрибог.
    pub fn sign_hash<R: Rng + CryptoRng>(
        &self,
        hash: &[u8],
        rng: &mut R,
    ) -> Result<Ec512Signature, ()> {
        unsafe {
            if hash.len() != EC_512_WIDTH {
                return Err(());
            }

            let cid = C::CURVE_ID;
            let mut seed = [0; EC_512_WIDTH];
            let mut inner = [0; EC_512_DOUBLE_WIDTH];

            loop {
                rng.fill_bytes(&mut seed);

                let res = sys::ec512_dsa_sign_hash(
                    cid,
                    self.as_ptr() as *mut _,
                    hash.as_ptr(),
                    seed.as_ptr(),
                    inner.as_mut_ptr(),
                );

                if res == 0 {
                    break;
                }
            }

            Ok(Ec512Signature { inner })
        }
    }

    /// Подписывает сообщение с помощью закрытого ключа
    pub fn sign<R: Rng + CryptoRng>(&self, msg: &[u8], rng: &mut R) -> Ec512Signature {
        unsafe {
            let cid = C::CURVE_ID;
            let mut seed = [0; EC_512_WIDTH];
            let mut inner = [0; EC_512_DOUBLE_WIDTH];

            loop {
                rng.fill_bytes(&mut seed);

                let res = sys::ec512_dsa_sign(
                    cid,
                    self.as_ptr() as *mut _,
                    msg.as_ptr(),
                    msg.len(),
                    seed.as_ptr(),
                    inner.as_mut_ptr(),
                );

                if res == 0 {
                    break;
                }
            }

            Ec512Signature { inner }
        }
    }

    /// Вырабатывает алгоритм выработки ключей обмена длиной 512 бит, используя 512-битные
    /// эллиптические кривые.
    ///
    /// Если `ukm` равен `None`, то используется значение по умолчанию равное единице. Если же
    /// `ukm` не `None`, то значение внутри `Some` не должно быть равным нулю по модулю простого
    /// числа эллиптической кривой. В случае передачи такого значения возвращается ошибка. Если
    /// используемый приватный или публичный ключ статичны (т.е. используются неэфемерные ключи),
    /// то рекомендуется использовать `ukm`, содержащий не менее 128 переменных бит.
    pub fn vko256(
        &self,
        other_pk: &Ec512PublicKey<C>,
        ukm: Option<[u8; EC_512_WIDTH]>,
    ) -> Result<[u8; EC_256_WIDTH], ()> {
        let cid = C::CURVE_ID;
        let mut out = MaybeUninit::<[u8; EC_256_WIDTH]>::uninit();

        let ukm = match ukm {
            None => core::ptr::null(),
            Some(ref v) => v as *const _,
        };

        unsafe {
            let res = sys::ec512_vko256(
                cid,
                self.as_ptr() as *mut _,
                other_pk.as_ptr() as *mut _,
                ukm,
                out.as_mut_ptr().cast(),
            );

            if res == 0 {
                Ok(out.assume_init())
            } else {
                Err(())
            }
        }
    }

    pub fn vko512(
        &self,
        other_pk: &Ec512PublicKey<C>,
        ukm: Option<[u8; EC_512_WIDTH]>,
    ) -> Result<[u8; EC_512_WIDTH], ()> {
        let cid = C::CURVE_ID;
        let mut out = MaybeUninit::<[u8; EC_512_WIDTH]>::uninit();

        let ukm = match ukm {
            None => core::ptr::null(),
            Some(ref v) => v as *const _,
        };

        unsafe {
            let res = sys::ec512_vko512(
                cid,
                self.as_ptr() as *mut _,
                other_pk.as_ptr() as *mut _,
                ukm,
                out.as_mut_ptr().cast(),
            );

            if res == 0 {
                Ok(out.assume_init())
            } else {
                Err(())
            }
        }
    }
}

impl<C: Ec512Curve> Drop for Ec512PrivateKey<C> {
    fn drop(&mut self) {
        unsafe {
            sys::ec512_private_key_drop(self.as_mut_ptr());
        }
    }
}

impl<C: Ec512Curve> Ec512PublicKey<C> {
    /// Десериализует открытый ключ из представления ГОСТ 34.10-2012 раздел 5.2
    pub fn from_bytes(bytes: &[u8]) -> Result<Self, ()> {
        unsafe {
            if bytes.len() != EC_512_DOUBLE_WIDTH {
                return Err(());
            }

            let cid = C::CURVE_ID;
            let mut public_key = Self::zeroed();

            let res =
                sys::ec512_public_key_from_bytes(cid, bytes.as_ptr(), public_key.as_mut_ptr());

            if res != 0 {
                return Err(());
            }

            Ok(public_key)
        }
    }

    /// Сериализует открытый ключ из внутреннего представления в представление ГОСТ 34.10-2012 раздел 5.2
    pub fn to_bytes(&self) -> [u8; EC_512_DOUBLE_WIDTH] {
        unsafe {
            let cid = C::CURVE_ID;
            let mut res = [0; EC_512_DOUBLE_WIDTH];

            // Ignore error because type system prevents us from this error
            let _res =
                sys::ec512_public_key_to_bytes(cid, self.as_ptr() as *mut _, res.as_mut_ptr());

            res
        }
    }

    /// Проверяет подпись хэша сообщения с помощью публичной ключа. По ГОСТу хэш сообщения должен
    /// быть вычислен с помощью функции Стрибог.
    pub fn verify_hash(&self, hash: &[u8], signature: &Ec512Signature) -> Result<(), ()> {
        unsafe {
            let cid = C::CURVE_ID;

            let res = sys::ec512_dsa_verify_hash(
                cid,
                self.as_ptr() as *mut _,
                hash.as_ptr(),
                signature.as_ptr(),
            );

            if res != 0 {
                return Err(());
            }

            Ok(())
        }
    }

    /// Проверяет подпись сообщения с помощью публичной ключа
    pub fn verify(&self, msg: &[u8], signature: &Ec512Signature) -> Result<(), ()> {
        unsafe {
            let cid = C::CURVE_ID;

            let res = sys::ec512_dsa_verify(
                cid,
                self.as_ptr() as *mut _,
                msg.as_ptr(),
                msg.len(),
                signature.as_ptr(),
            );

            if res != 0 {
                return Err(());
            }

            Ok(())
        }
    }
}
