#include "catch_amalgamated.hpp"
#include "gostcrypto.hpp"
#include "common.hpp"

template <typename CfbType, size_t NonceSize>
void cfb_test()
{
    const auto key = random_array<SYMMETRIC_KEY_SIZE>();
    const auto nonce = random_array<NonceSize>();

    constexpr size_t buf_len = 1024;

    CfbType cfb(key, nonce);
    CfbType cfb_copy(key, nonce);

    std::vector<uint8_t> buf(buf_len);
    auto expected = buf;

    cfb.encrypt(buf);
    
    REQUIRE(buf != expected);
    
    cfb_copy.decrypt(buf);

    REQUIRE(buf.size() == expected.size());
    
    REQUIRE(buf == expected);
}

template <typename CtrType, size_t NonceSize, bool is_seekable = true>
void ctr_test()
{
    const auto key = random_array<SYMMETRIC_KEY_SIZE>();
    const auto nonce = random_array<NonceSize>();

    constexpr size_t buf_len = 1024;

    SECTION("copy and apply_keystream")
    {
        CtrType ctr(key, nonce);
        CtrType ctr_copy = ctr;

        std::vector<uint8_t> buf(buf_len);
        auto buf2 = buf;
        ctr.apply_keystream(buf);
        ctr_copy.apply_keystream(buf2);
        REQUIRE(buf == buf2);
    }

    SECTION("seek and apply_keystream")
    {
        CtrType ctr(key, nonce);

        std::vector<uint8_t> buf(buf_len);
        auto buf2 = buf;
        ctr.apply_keystream(buf);
        ctr.seek(0);
        ctr.apply_keystream(buf2);

        REQUIRE(buf == buf2);

        uint64_t seek_pos = 42;
        ctr.seek(seek_pos);

        REQUIRE(seek_pos == ctr.get_pos());
    }
}

template <typename OfbType, size_t NonceSize>
void ofb_test()
{
    const auto key = random_array<SYMMETRIC_KEY_SIZE>();
    const auto nonce = random_array<NonceSize>();

    constexpr size_t buf_len = 1024;

    OfbType ofb(key, nonce);

    OfbType ofb_copy = ofb;

    std::vector<uint8_t> buf(buf_len);
    auto buf2 = buf;
    ofb.apply_keystream(buf);
    ofb_copy.apply_keystream(buf2);

    REQUIRE(buf == buf2);
}

TEST_CASE("MagmaCfb")
{
    cfb_test<MagmaCfb_t, MAGMA_CFB_NONCE_SIZE>();
}

TEST_CASE("KuznyechikCfb")
{
    cfb_test<KuznyechikCfb_t, KUZNYECHIK_CFB_NONCE_SIZE>();
}

TEST_CASE("MagmaOfb")
{
    ofb_test<MagmaOfb_t, MAGMA_OFB_NONCE_SIZE>();
}

TEST_CASE("KuznyechikOfb")
{
    ofb_test<KuznyechikOfb_t, KUZNYECHIK_OFB_NONCE_SIZE>();
}

TEST_CASE("KuznyechikCtr")
{
    ctr_test<KuznyechikCtr_t, KUZNYECHIK_CTR_NONCE_SIZE>();
}

TEST_CASE("MagmaCtr")
{
    ctr_test<MagmaCtr_t, MAGMA_CTR_NONCE_SIZE>();
}
