#include "catch_amalgamated.hpp"
#include "gostcrypto.hpp"
#include "common.hpp"

template <Gost3410_2012_256 CurveID>
void vko256_test()
{
    intptr_t error_code;

    Ec256PrivateKey_t<CurveID> sk1 = gen_private_key<Ec256PrivateKey_t<CurveID>, EC_256_WIDTH>(error_code);
    REQUIRE(error_code == 0);
    auto pk1 = sk1.get_public_key();

    Ec256PrivateKey_t<CurveID> sk2 = gen_private_key<Ec256PrivateKey_t<CurveID>, EC_256_WIDTH>(error_code);
    REQUIRE(error_code == 0);
    auto pk2 = sk2.get_public_key();

    SECTION("NULL ukm")
    {
        const auto key1 = sk1.vko256(pk2, error_code);
        REQUIRE(error_code == 0);

        const auto key2 = sk2.vko256(pk1, error_code);
        REQUIRE(error_code == 0);

        REQUIRE(key1 == key2);

        const std::array<uint8_t, EC_256_WIDTH> zeroed_key = {};
        REQUIRE(key1 != zeroed_key);
        REQUIRE(key2 != zeroed_key);
    }

    SECTION("rand ukm")
    {
        const auto ukm = random_array<EC_256_WIDTH>();

        const auto key1 = sk1.vko256(pk2, error_code, ukm.data());
        REQUIRE(error_code == 0);

        const auto key2 = sk2.vko256(pk1, error_code, ukm.data());
        REQUIRE(error_code == 0);

        REQUIRE(key1 == key2);

        const std::array<uint8_t, EC_256_WIDTH> zeroed_key = {};
        REQUIRE(key1 != zeroed_key);
        REQUIRE(key2 != zeroed_key);
    }

    SECTION("bad ukm")
    {
        const std::array<uint8_t, EC_256_WIDTH> zeroed_ukm = {};

        sk1.vko256(pk2, error_code, zeroed_ukm.data());
        REQUIRE(error_code != 0);

        sk2.vko256(pk1, error_code, zeroed_ukm.data());
        REQUIRE(error_code != 0);
    }
}

template <Gost3410_2012_512 CurveID>
void vko512_test()
{
    intptr_t error_code;

    Ec512PrivateKey_t<CurveID> sk1 = gen_private_key<Ec512PrivateKey_t<CurveID>, EC_512_WIDTH>(error_code);
    REQUIRE(error_code == 0);
    auto pk1 = sk1.get_public_key();

    Ec512PrivateKey_t<CurveID> sk2 = gen_private_key<Ec512PrivateKey_t<CurveID>, EC_512_WIDTH>(error_code);
    REQUIRE(error_code == 0);
    auto pk2 = sk2.get_public_key();

    // SECTION("vko256")
    {
        SECTION("NULL ukm")
        {
            const auto key1 = sk1.vko256(pk2, error_code);
            REQUIRE(error_code == 0);

            const auto key2 = sk2.vko256(pk1, error_code);
            REQUIRE(error_code == 0);

            REQUIRE(key1 == key2);

            const std::array<uint8_t, EC_256_WIDTH> zeroed_key = {};
            REQUIRE(key1 != zeroed_key);
            REQUIRE(key2 != zeroed_key);
        }

        SECTION("rand ukm")
        {
            const auto ukm = random_array<EC_512_WIDTH>();

            const auto key1 = sk1.vko256(pk2, error_code, ukm.data());
            REQUIRE(error_code == 0);

            const auto key2 = sk2.vko256(pk1, error_code, ukm.data());
            REQUIRE(error_code == 0);

            REQUIRE(key1 == key2);

            const std::array<uint8_t, EC_256_WIDTH> zeroed_key = {};
            REQUIRE(key1 != zeroed_key);
            REQUIRE(key2 != zeroed_key);
        }

        SECTION("bad ukm")
        {
            const std::array<uint8_t, EC_512_WIDTH> zeroed_ukm = {};

            sk1.vko256(pk2, error_code, zeroed_ukm.data());
            REQUIRE(error_code != 0);

            sk2.vko256(pk1, error_code, zeroed_ukm.data());
            REQUIRE(error_code != 0);
        }
    }

    // SECTION("vko512")
    {
        SECTION("NULL ukm")
        {
            const auto key1 = sk1.vko512(pk2, error_code);
            REQUIRE(error_code == 0);

            const auto key2 = sk2.vko512(pk1, error_code);
            REQUIRE(error_code == 0);

            REQUIRE(key1 == key2);

            const std::array<uint8_t, EC_512_WIDTH> zeroed_key = {};
            REQUIRE(key1 != zeroed_key);
            REQUIRE(key2 != zeroed_key);
        }

        SECTION("rand ukm")
        {
            const auto ukm = random_array<EC_512_WIDTH>();

            const auto key1 = sk1.vko512(pk2, error_code, ukm.data());
            REQUIRE(error_code == 0);

            const auto key2 = sk2.vko512(pk1, error_code, ukm.data());
            REQUIRE(error_code == 0);

            REQUIRE(key1 == key2);

            const std::array<uint8_t, EC_512_WIDTH> zeroed_key = {};
            REQUIRE(key1 != zeroed_key);
            REQUIRE(key2 != zeroed_key);
        }

        SECTION("bad ukm")
        {
            const std::array<uint8_t, EC_512_WIDTH> zeroed_ukm = {};

            sk1.vko512(pk2, error_code, zeroed_ukm.data());
            REQUIRE(error_code != 0);

            sk2.vko512(pk1, error_code, zeroed_ukm.data());
            REQUIRE(error_code != 0);
        }
    }
}

TEST_CASE("Ec256::A VKO")
{
    vko256_test<Gost3410_2012_256::A>();
}

TEST_CASE("Ec256::B VKO")
{
    vko256_test<Gost3410_2012_256::B>();
}

TEST_CASE("Ec256::C VKO")
{
    vko256_test<Gost3410_2012_256::C>();
}

TEST_CASE("Ec256::D VKO")
{
    vko256_test<Gost3410_2012_256::D>();
}

TEST_CASE("Ec512::A VKO")
{
    vko512_test<Gost3410_2012_512::A>();
}

TEST_CASE("Ec512::B VKO")
{
    vko512_test<Gost3410_2012_512::B>();
}

TEST_CASE("Ec512::C VKO")
{
    vko512_test<Gost3410_2012_512::C>();
}
